"""Logitech Unifying mouse tool

This utility provides multiple features targeting Logitech Unifying mice:
- mouse movement and click logging
- interactive mouse and click injection (taking control over a wireless mouse)
- mouse and click injection from stdin
"""
import sys
import time
import logging

# Whad dependencies
from pynput import mouse
from whad.cli.app import CommandLineSink, run_app
from whad.unifying.connector import Keyboard, Keylogger, ESBAddress
from whad.unifying.stack.constants import ClickType
from whad.esb.exceptions import InvalidESBAddressException
from whad.exceptions import WhadDeviceNotFound

# Logging
logger = logging.getLogger(__name__)

TOOL_DESCRIPTION="""This tool synchronizes with an existing keyboard and logs keypresses or
inject specific payloads to the associated dongle."""

class DuckyScriptError(Exception):
    """DuckyScript error.
    """

    def __init__(self, message):
        super().__init__()
        self.__message = message

    def __str__(self):
        return f"DuckyScriptError(\"{self.__message}\")"

    def __repr__(self):
        return str(self)
    

class DuckyScriptRunner(object):
    """Ducky Script (tm) parser and runner.

    This class processes a DuckyScript in text form (not compiled) and executes
    it through python function instrumentation. This is a quick and dirty runner
    compatible with DuckyScript v1. 
    """

    # Commands format
    COMMANDS = {
        # Discard comments
        "//": None,
        "REM": None,

        # Delays
        "DEFAULTDELAY": "i",
        "DEFAULTCHARDELAY": "i",
        "CHARJITTER":  "i",
        "DELAY": "i",

        # String lines and blocks
        "STRING": "s",
        "STRINGLN": "s",
        "STRINGLN_BLOCK": "",
        "END_STRINGLN": "",
        "STRING_BLOCK": "",
        "END_STRING": "",

        # Keys
        "KEYDOWN": "k",
        "KEYUP": "k",

        # Loops
        "REPEAT": "i",
    }

    def __init__(self, file_path : str, connector: Keyboard):
        """Initialize ducky script parser.

        :param file_path: DuckyScript path
        :type file_path: str
        """
        self.__script_path = file_path
        self.__connector = connector
        self.__instructions = []

        # DuckyScript execution parameters
        self.__default_delay = 18
        self.__default_char_delay = 18
        self.__char_jitter = False

    def parse_operand(self, token: str, format: str) -> dict:
        """Transform an operand to an abstract notation, including
        its expected format.

        :param token: operand token
        :type token: str
        :param format: expected format
        :type format: str
        :return: dict representing this operand
        :rtype: dict
        """
        return {
            "value": token,
            "format": format,
        }

    def parse_operands(self, tokens: list, format: str) -> list:
        """Parse a set of operands from tokens and following
        expected operands format.

        :param tokens: list of operands as tokens
        :type tokens: list
        :param format: list of operands types
        :type format: str
        :return: list of parsed operands
        :rtype: list
        """
        parsed_operands = []
        
        # if no format provided, discard
        if format is None:
            return []

        # If we expect a single text string, all the tokens following the
        # instruction are part of the string. To handle this specific case,
        # we update the format string to take into account all remaining
        # tokens.
        if format.strip() == "s":
            format = "s"*len(tokens)
        elif len(tokens) != len(format):
            raise DuckyScriptError(f"Invalid format for instruction {tokens[0]}")

        # Try to convert token to expected format
        for i, token in enumerate(tokens):
            parsed_operands.append(self.parse_operand(token.strip(), format[i]))
        
        return parsed_operands


    def parse_line(self, line: str) -> dict:
        """This function parses a DuckyScript line and turns it into an abstract
        representation.

        :param line: line to parse
        :type line: str
        :return: abstract representation for this line
        :rtype: dict
        """
        # Trim line
        line = line.strip()

        # Split line
        tokens = line.split(' ')

        # Translate tokens to instructions
        inst = tokens[0].upper()

        # Parse command
        if inst in DuckyScriptRunner.COMMANDS:
            operands = self.parse_operands(
                tokens[1:],
                DuckyScriptRunner.COMMANDS[inst]
            )
        
            # Return instruction
            return {
                "inst": inst,
                "operands": operands
            }
        else:
            return {
                "data": line 
            }

    def to_multi_string(self, lines: list, last_enter=True) -> list:
        """Convert a series of lines to a series of typed text followed
        by an ENTER key after each line.

        :param lines: List of lines to interpret as text
        :type lines: list
        :param last_enter: If set to True, will include a press on ENTER key
                           for the last text line
        :type last_enter: bool
        :return: list of asbtract instructions
        """
        output = []
        for line in lines:
            output.append({
                "func": "send_text",
                "args":[line["data"]]
            })
            output.append({
                "func": "send_keys",
                "args": ["ENTER"]
            })

        # Remove last ENTER key press if required
        if not last_enter:
            output = output[:-1]

        return output

    def to_instructions(self, lines: list) -> list:
        """Convert lines to instructions.

        :param lines: Lines to convert to instructions
        :type lines: list
        :return: List of instructions
        """
        block = []

        # Loop on lines
        i = 0
        while i < len(lines):
            # Get current line
            line = lines[i]

            # If current line contains an instruction:
            if "inst" in line:
                # Don't process comments
                if line["inst"] in ["REM", "//"]:
                    pass
                elif line["inst"] == "DEFAULTDELAY":
                    block.append({
                        "func": "set_delay",
                        "args": line["operands"]
                    })
                elif line["inst"] == "DEFAULTCHARDELAY":
                    block.append({
                        "func": "set_char_delay",
                        "args": line["operands"]
                    })
                elif line["inst"] == "CHARJITTER":
                    block.append({
                        "func": "char_jitter",
                        "args": line["operands"]
                    })
                elif line["inst"] == "DELAY":
                    block.append({
                        "func": "delay",
                        "args": line["operands"]
                    })
                elif line["inst"] == "STRING":
                    # Combine operands to rebuild our string
                    op_string = []
                    for str_bit in line["operands"]:
                        if str_bit["format"] != "s":
                            raise DuckyScriptError("Syntax error line {i}")
                        op_string.append(str_bit["value"])
                    op_string = " ".join(op_string)

                    block.append({
                        "func": "send_text",
                        "args": [op_string]
                    })
                elif line["inst"] == "STRINGLN":
                    # Combine operands to rebuild our string
                    op_string = []
                    for str_bit in line["operands"]:
                        if str_bit["format"] != "s":
                            raise DuckyScriptError("Syntax error line {i}")
                        op_string.append(str_bit["value"])
                    op_string = " ".join(op_string)

                    block.append({
                        "func": "send_textline",
                        "args": [op_string]
                    })

                elif line["inst"] == "STRING_BLOCK":
                    # Pile up the following lines until we meet a "END_STRING"
                    # instruction.
                    j = i + 1
                    found_end = False
                    while j < len(lines):
                        if "inst" in lines[j] and lines[j]["inst"] == "END_STRING":
                            block.append({
                                "func": "execute",
                                "args": [self.to_multi_string(lines[i+1:j], last_enter=False)]
                            })
                            found_end = True
                            i = j + 1
                            break
                        else:
                            j += 1
                    
                    # Raise an error if we did not find the END_STRING command
                    if not found_end:
                        raise DuckyScriptError(f"Unfinished STRING_BLOCK, line {i}")
                elif line["inst"] == "STRINGLN_BLOCK":
                    # Pile up the following lines until we meet a "END_STRING"
                    # instruction.
                    j = i + 1
                    found_end = False
                    while i < len(lines):
                        if "inst" in lines[j] and lines[j]["inst"] == "END_STRINGLN":
                            block.append({
                                "func": "execute",
                                "args": [self.to_multi_string(lines[i+1:j])]
                            })
                            found_end = True
                            i = j + 1
                            break
                        else:
                            j += 1

                    # Raise an error if we did not find the END_STRING command
                    if not found_end:
                        raise DuckyScriptError(f"Unfinished STRINGLN_BLOCK, line {i}")
                elif line["inst"] == "REPEAT":
                    # Build a repeat instruction based on loop
                    block.append({
                        "func": "do_repeat",
                        "args": [block[-1], line["operands"][0]]
                    })

            # If line is data and we are not in a block
            elif "data" in line:
                # Parse data
                block.append({
                    "func": "send_keys",
                    "args": [line["data"]]
                })
            else:
                raise DuckyScriptError("Syntax error on line {i}")

            # Go next line
            i += 1

        # Execute our translated script
        return block

    def evaluate_expr(self, expr: dict):
        """Evaluate expression.

        This function only evaluates integer for now, but in a near future will
        also accept variables and basic operations.
        """
        expr_format = expr["format"]
        value = expr["value"]

        # Evaluate integer
        if expr_format == "i":
            return int(value)
        else:
            return None

    def execute(self, block: list):
        """Execute a block of asbtract instructions.

        :param block: List of abstract instructions
        :type block: list
        """
        for func_call in block:
            if "func" in func_call and "args" in func_call:
                func_name = func_call["func"]
                function = getattr(self, func_name)
                if function is not None:
                    # Call function with provided args
                    function(*func_call["args"])
                else:
                    raise DuckyScriptError(f"Cannot execute native method {func_name}.")
            else:
                raise DuckyScriptError("Wrong native function call for object {func_call}.")
            
            # Wait the specified delay between lines
            time.sleep(self.__default_delay/1000.)

    def run(self):
        """Run the provided DuckyScript using a Keyboard connector.

        This function parses a DuckyScript, transforms it into an abstract
        representation and executes it against the target device.
        """
        # Load script from file and tokenize
        try:
            self.__lines = []

            # First, parse all lines
            line_counter = 1
            with open(self.__script_path, 'r', encoding='utf-8') as script:
                for line in script:
                    # Trim line and parse
                    self.__lines.append(self.parse_line(line))
                    line_counter += 1

            # Convert lines to scripted calls
            script = self.to_instructions(self.__lines)

            # Execute script
            self.execute(script)
        except IOError:
            logger.debug("I/O Error while reading script %s", self.__script_path)
        except DuckyScriptError as exc:
            print(line)
            print(exc)
            logger.error("DuckyScript parsing error at line %d", line_counter)

    def do_repeat(self, inst, operand):
        """Implement the REPEAT operation of DuckyScript.

        This function evaluates the expression passed as operand, representing
        the number of iterations to perform, and execute the corresponding
        instruction as many times as required.

        :param inst: Abstract instruction to execute
        :type inst: dict
        :param operand: Abstract operand
        :type operand: dict
        """
        nb = self.evaluate_expr(operand)
        for _ in range(nb):
            self.execute([inst])

    def char_jitter(self, value):
        """Set character jitter.

        Character jitter is used to add random waiting times between key
        presses in order to mimick human behavior. For now, this feature is
        implemented but not effective as the underlying stack does not support
        it yet.

        :param value: jitter value to set
        :type value: dict
        """
        self.__char_jitter = self.evaluate_expr(value)

    def send_text(self, text: str):
        """Send text (without hitting ENTER) to the target Unifying dongle.

        :param text: Text to send
        :type text: str
        """
        print(text)
        self.__connector.send_text(text)

    def send_textline(self, text):
        """Send text (including final ENTER press) to the target Unifying
        dongle.

        :param text: Text to send
        :type text: str
        """
        self.__connector.send_text(text)
        self.__connector.send_key("ENTER")

    def send_keys(self, keys):
        """Send keys combination to the target keyboard dongle.

        :param keys: Keys combination as described in DuckyScript
        :type keys: str
        """
        gui = False
        shift = False
        ctrl = False
        alt = False

        # Split keys and extract special keys (gui, shift, control, alt)
        other_keys = []
        keys = keys.split(' ')
        for key in keys:
            key = key.strip()
            if key == "WINDOWS":
                gui = True
            elif key in ["SHIFT", "RSHIFT"]:
                shift = True
            elif key in ["CONTROL", "CTRL", "RCTRL", "RCONTROL"]:
                ctrl = True
            elif key in ["ALT", "RALT"]:
                alt = True
            else:
                other_keys.append(key)
        if len(other_keys) > 0:
            key = other_keys[0]
        else:
            key = ""

        # Send key through connector
        self.__connector.send_key(key, gui=gui, alt=alt, shift=shift, ctrl=ctrl)
        
    def set_delay(self, delay):
        """Set default delay.

        :param delay: Abstract operand that specifies the delay (in milliseconds)
        :type delay: dict
        """
        delay = self.evaluate_expr(delay)
        self.__default_delay = delay

    def set_char_delay(self, delay):
        """Set default char delay.

        :param delay: Abstract operand that specifies the delay (in milliseconds)
        :type delay: dict
        """
        delay = self.evaluate_expr(delay)
        self.__default_char_delay = delay

    def delay(self, delay):
        """Wait for a specific delay in ms.

        :param delay: Abstract operand that specifies the delay (in milliseconds)
        :type delay: dict
        """
        delay = self.evaluate_expr(delay)
        time.sleep(delay/1000.)


class UniKeyboardApp(CommandLineSink):
    """Logitech Unifying keyboard CLI app
    """

    mouse_coords = None

    def __init__(self):
        """Application uses an interface and has no commands.
        """
        super().__init__(
            description=TOOL_DESCRIPTION,
            interface=True,
            commands=False
        )

        # Target device address option
        self.add_argument(
            "-a",
            "--address",
            metavar='ADDRESS',
            dest='address',
            default=None,
            help="Target keyboard with address ADDRESS"
        )

        # Payload to inject
        self.add_argument(
            "-d",
            "--ducky",
            metavar="DUCKYSCRIPT",
            dest="ducky",
            type=str,
            help="DuckyScript to execute"
        )

        # Payload to inject
        self.add_argument(
            "-p",
            "--payload",
            metavar="PAYLOAD",
            dest="payload",
            type=str,
            help="Payload to inject"
        )

        # Keymap
        self.add_argument(
            "-l",
            "--locale",
            metavar="LOCALE",
            dest="locale",
            type=str,
            default="us",
            help="Use locale LOCALE (default: us)"
        )

        # Encryption key
        self.add_argument(
            "-k",
            "--key",
            metavar="ENCKEY",
            dest="key",
            type=str,
            help="Set encryption key (hex), will enable encryption/decryption if set"
        )

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                # If an address is not provided, we cannot go further
                if self.args.address is None:
                    # Start scanning
                    self.error("You must specify a target keyboard address with -a/--address")
                else:
                    # If an address is provided, we check its format and start
                    # sniffing packets from this device

                    # Parse encryption key if provided
                    if self.args.key is not None:
                        try:
                            enc_key = bytes.fromhex(self.args.key)
                        except ValueError:
                            logger.debug("encryption key '%s' cannot be decoded.", self.args.key)
                            logger.debug("Make sure the encryption key is 128-bit long and in correct hex.")
                            enc_key = None
                    else:
                        enc_key = None

                    try:
                        # Check address format
                        addr = ESBAddress(self.args.address)
                        
                        # If -p/--payload option is set, sync and replicate our mouse moves
                        if self.args.payload is not None:
                            self.send_payload(addr, self.args.payload, self.args.locale, key=enc_key)
                        elif self.args.ducky is not None:
                            self.send_ducky(addr, self.args.ducky, self.args.locale, key=enc_key)
                        else:
                            # If stdin is piped, then we expect some mouse moves and clicks
                            # coming from stdin
                            if self.is_stdin_piped():
                                self.send_stdin(addr, self.args.locale, key=enc_key)
                            else:
                                
                                # Log keyboard events
                                self.log_keyboard(addr, self.args.locale, key=enc_key)
                    except InvalidESBAddressException:
                        # Invalid device address
                        self.error('Target address does not match the expected format !')
            else:
                # Missing interface.
                self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt:
            self.warning('wuni-keyboard stopped (CTL-C)')

        # Launch post-run tasks
        self.post_run()

    def send_payload(self, address: ESBAddress, payload: str, locale: str, key: bytes = None):
        """Send payload to target keyboard.
        """
        # Connect to target device and performs discovery
        try:
            # Create our keyboard connector and set target address
            connector = Keyboard(self.interface)
            connector.address = str(address)

            # Set connector stack's locale
            connector.stack.get_layer('app').locale = locale

            # Set encryption key and counter if key is provided
            if key is not None:
                connector.key = key
                connector.aes_counter = 0

            # Start connector (enable keyboard mode)
            connector.start()
            
            # Synchronize with the dongle
            connector.synchronize()

            # Lock the dongle on the current channel
            connector.lock()

            # Send text payload followed by key
            connector.send_text(payload)
            connector.send_key("ENTER")

            # Stop connector
            connector.unlock()
            connector.stop()
        except (KeyboardInterrupt, SystemExit):
            connector.stop()

    def send_ducky(self, address: ESBAddress, ducky_script: str, locale: str, key: bytes = None):
        """Send ducky script
        """
        # Create our keyboard connector and set target address
        connector = Keyboard(self.interface)
        runner = DuckyScriptRunner(ducky_script, connector)
        connector.address = str(address)

        # Set connector stack's locale
        connector.stack.get_layer('app').locale = locale

        # Set encryption key and counter if key is provided
        if key is not None:
            connector.key = key
            connector.aes_counter = 0

        # Start connector (enable keyboard mode)
        connector.start()
        
        # Synchronize with the dongle
        connector.synchronize()

        # Lock the dongle on the current channel
        connector.lock()
        #time.sleep(.5)

        runner.run()

        connector.unlock()
        connector.stop()

    def send_stdin(self, address: ESBAddress, locale: str, key: bytes = None):
        """Read lines from stdin and send text (including ENTER) to target
        keyboard.
        """

        try:
            # Create our keyboard connector and set target address
            connector = Keyboard(self.interface)
            connector.address = str(address)

            # Set connector stack's locale
            connector.stack.get_layer('app').locale = locale

            # Set encryption key and counter if key is provided
            if key is not None:
                connector.key = key
                connector.aes_counter = 0

            # Start connector (enable keyboard mode)
            connector.start()

            # Synchronize with the dongle
            connector.synchronize()

            # Lock the dongle on the current channel
            connector.lock()

            # Wait for the dongle to be ready for injection
            time.sleep(.5)

            # Send data coming fron stdin (strip newlines and press ENTER instead)
            for line in sys.stdin:
                connector.send_text(line.strip())
                connector.send_key("ENTER")

            # Close connector
            connector.unlock()
            connector.close()

        except (KeyboardInterrupt, SystemExit):
            connector.stop()

    def log_keyboard(self, address: ESBAddress, locale: str, key: bytes = None):
        """Log keyboard presses (if not encrypted)
        """
        # Connect to target device and performs discovery
        try:
            connector = Keylogger(self.interface)
            connector.address = str(address)
            connector.scanning = True
            connector.locale = locale

            # Set encryption key if provided
            if key is not None:
                connector.decrypt = True
                connector.add_key(key)

            # Start logging keyboard events
            connector.start()
            out = ""
            for i in connector.stream():
                out += i
                print(out)

        except (KeyboardInterrupt, SystemExit):
            connector.close()

        except WhadDeviceNotFound:
            print('[e] Device not found')
            exit(1)

def wuni_kb_main():
    """Logitech Unifying keyboard tool main routine.
    """
    app = UniKeyboardApp()
    run_app(app)



"""Bluetooth Low Energy packet injection module.

This module provides the :class:`whad.ble.connector.injection.Injector` class
to perform packet injection.
"""
from time import sleep

from scapy.layers.bluetooth4LE import BTLE, BTLE_ADV, BTLE_DATA

from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter
from whad.hub.ble import Injected, Direction as BleDirection
from whad.ble.sniffing import SynchronizedConnection
from whad.ble.injecting import InjectionConfiguration

from ..exceptions import NotSynchronized
from .base import BLE


class Injector(BLE):
    """BLE injecion connector.
    """

    def __init__(self, device, connection=None):
        super().__init__(device)
        self.__connection = connection
        self.__synchronized = connection is not None

        # Check if device accepts injection
        if not self.can_inject():
            raise UnsupportedCapability("Inject")

        self.__configuration = InjectionConfiguration()


    def _enable_configuration(self):
        """Enable configuration for injection.
        """
        if self.__configuration.synchronize:
            if self.__configuration.active_connection is not None:
                if not self.can_sniff_active_connection():
                    raise UnsupportedCapability("Sniff")
                # if active connection has been provided, synchronize according to parameters
                access_address = self.__configuration.active_connection.access_address
                crc_init = self.__configuration.active_connection.crc_init
                channel_map = self.__configuration.active_connection.channel_map
                hop_interval = self.__configuration.active_connection.hop_interval
                hop_increment = self.__configuration.active_connection.hop_increment
                self.sniff_active_connection(access_address, crc_init, channel_map,
                                             hop_interval, hop_increment)
            else:
                if not self.can_sniff_new_connection():
                    raise UnsupportedCapability("Sniff")
                if self.__configuration.channel is not None:
                    channel = self.__configuration.channel
                else:
                    channel = 37 # fallback
                self.sniff_new_connection(
                    channel=channel,
                    show_advertisements=False,
                    show_empty_packets=False,
                    bd_address=self.__configuration.filter
                )
            self.start()
            while not self.__synchronized:
                sleep(1)
            return True
        return True

    def configure(self, active_connection=None, synchronize=False, channel=37,
                  bd_filter="FF:FF:FF:FF:FF:FF"):
        """Configure this connector to target an active connection.
        """
        self.stop()
        self.__configuration.active_connection = active_connection
        self.__configuration.synchonize = synchronize
        self.__configuration.channel = channel
        self.__configuration.filter = bd_filter
        self._enable_configuration()


    @property
    def configuration(self):
        """Current injection configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_configuration()

    def inject(self, packet):
        """Inject packet in current connection.
        """
        if self.__configuration.raw:
            return self.raw_inject(packet)
        if self.__configuration.inject_to_slave:
            return self.inject_to_slave(packet)
        if self.__configuration.inject_to_master:
            return self.inject_to_master(packet)

        # Cannot inject
        return False

    def raw_inject(self, packet):
        """
        Inject a raw packet directly, according to the channel provided in
        configuration or metadata.
        """
        if BTLE in packet:
            access_address = packet.access_addr
        elif BTLE_ADV in packet:
            access_address = 0x8e89bed6
        elif BTLE_DATA in packet:
            if self.__connection is not None:
                access_address = self.__connection.access_address
            else:
                access_address = 0x11223344 # default value
        else:
            # Fail, no access address found
            return None

        if self.__configuration.channel is not None:
            channel = self.__configuration.channel
        if hasattr(packet, "metadata") and hasattr(packet.metadata, "channel"):
            channel = packet.metadata.channel
        else:
            channel = 37 # fallback to channel 37

        return self.send_pdu(packet, access_address=access_address, conn_handle=channel,
                             direction=BleDirection.UNKNOWN)

    def inject_to_slave(self, packet):
        """Inject packet in existing connection, targeting the advertiser.
        """
        if self.__connection is not None:
            self.send_pdu(packet, access_address=self.__connection.access_address,
                          direction=BleDirection.INJECTION_TO_SLAVE)
            message = self.wait_for_message(keep=message_filter(Injected))
            return (message.success, message.injection_attempts)

        # Not synchronized
        raise NotSynchronized()

    def inject_to_master(self, packet):
        """Inject packet in existing connection, targeting the initiator.
        """
        if self.__connection is not None:
            self.send_pdu(packet, access_address=self.__connection.access_address,
                          direction=BleDirection.INJECTION_TO_MASTER)
            message = self.wait_for_message(keep=message_filter(Injected))
            return (message.success, message.injection_attempts)

        # Not synchronized
        raise NotSynchronized()


    def on_synchronized(self, access_address=None, crc_init=None, hop_increment=None,
                        hop_interval=None, channel_map=None):
        self.__connection = SynchronizedConnection(
            access_address = access_address,
            crc_init = crc_init,
            hop_increment = hop_increment,
            hop_interval = hop_interval,
            channel_map = channel_map
        )
        self.__synchronized = True

    def on_desynchronized(self, access_address=None):
        if access_address == self.__connection.access_address:
            self.__synchronized = False
            self.__connection = None



from whad.dot15d4.exceptions import Dot15d4TimeoutException
from functools import wraps
from time import time
from queue import Queue, Empty
from inspect import signature
import logging

class Dot15d4Service:
    """
    This class represents a 802.15.4 service, connected to a manager (e.g., WHAD generic stack Layer).
    It also implements a set of decorators to identify wireless stack mechanisms type according to
    802.15.4 specification (e.g., request, response, indication).
    """
    def __init__(self, manager, name=None, timeout_exception_class=Dot15d4TimeoutException):
        self._logger = logging.getLogger(self.__module__)
        self._manager = manager
        self._name = name if name is not None else self.__class__.__name__
        self._timeout_exception_class = timeout_exception_class
        self._queue = Queue()

    @property
    def name(self):
        """
        Returns the associated name.
        """
        return self._name

    @property
    def manager(self):
        """
        Returns the associated manager.
        """
        return self._manager

    @manager.setter
    def manager(self, manager):
        """
        Set the associated manager.
        """
        self._manager = manager

    @property
    def database(self):
        """
        Alias to facilitate the manipulation of manager database.
        """
        return self._manager.database

    def add_packet_to_queue(self, packet):
        """
        Add an incoming packet to the queue.
        """
        self._queue.put(packet, block=True, timeout=None)

    def wait_for_packet(self, packet_filter, timeout=1.0):
        """Wait for a specific message type or error, other messages are dropped

        :param type packet_filter: Filtering lambda
        :param float timeout: Timeout value (default: 1 second)
        """
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self._queue.get(block=False,timeout=0.1)
                if packet_filter(msg):
                    return msg
            except Empty:
                pass
        raise self._timeout_exception_class()


    # Services primitives decorator
    def request(request_name):
        """ Mark the associated method as a "Request".
        """
        def _request(func):
            @wraps(func)
            def request_decorator(*args, **kwargs):
                self = args[0]
                self._logger.info("[%s] %s request (%s,%s)", args[0]._name, request_name, str(args[1:]),str(kwargs))
                result = func(*args, **kwargs)
                self._logger.info("[%s] %s confirm (%s)", args[0]._name, request_name, str(result))
                return result
            return request_decorator
        return _request


    def response(response_name):
        """ Mark the associated method as a "Response".
        """
        def _response(func):
            @wraps(func)
            def response_decorator(*args, **kwargs):
                self = args[0]
                self._logger.info("[%s] %s response (%s,%s)", args[0]._name, response_name, str(args[1:]),str(kwargs))
                result = func(*args, **kwargs)
                return result
            return response_decorator
        return _response


    def indication(indication_name):
        """ Mark the associated method as an "Indication".
        """
        def _indication(func):
            @wraps(func)
            def indication_decorator(*args, **kwargs):
                self = args[0]
                pdu, parameters = func(*args, **kwargs)

                if self._manager.upper_layer is None:
                    self._logger.warning("[%s] %s indication not transmitted, no upper layer !", self._name, indication_name)
                    return None

                upper = self._manager.upper_layer.alias
                callback_kwargs = {"tag":indication_name}
                callback_kwargs.update(parameters)

                self._logger.info("[%s] %s indication (%s,%s)", self._name, indication_name, str(pdu), str(callback_kwargs))
                return_value = self._manager.send(upper, pdu, **callback_kwargs)
                return return_value

            return indication_decorator
        return _indication



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import sys
import logging
from typing import List, Tuple

from scapy.layers.all import *
from scapy.packet import Packet
from scapy.config import conf
from scapy.themes import BrightTheme

from whad.scapy.layers import *
from whad.cli.app import CommandLineApp, run_app
from whad.device.unix import  UnixConnector
from whad.hub import ProtocolHub

logger = logging.getLogger(__name__)

class WhadExtractApp(CommandLineApp):
    """
    Main `wextract` CLI application class.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD extraction tool',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_STANDARD
        )

        self.add_argument(
            'extractor',
            help='extractor to evaluate',
            nargs='*'
        )

        self.add_argument(
            '-d',
            '-D',
            '--delimiter',
            dest='delimiter',
            default=" ",
            help='delimiter between extractor'
        )

        self.add_argument(
            '-l',
            '--load',
            dest='loadables',
            default=None,
            action="append",
            help='load Scapy packet definitions from external Python file'
        )

        self.add_argument(
            '-x',
            '--exceptions',
            dest="show_exc",
            action="store_true",
            default=False,
            help="Show exceptions raised when extracting requested values (debug)"
        )

    def build_extractors(self) -> List[Tuple[str, callable]]:
        """Build extractors based on provided arguments.

        :rtype: list
        :return: list of extractors
        """
        extractor_template = "lambda p : %s"
        extractors = []
        for extractor in self.args.extractor:
            if "packet." in extractor:
                extractor.replace("packet.", "p.")
            elif "pkt." in extractor:
                extractor.replace("pkt.", "p.")
            extractors.append((extractor, eval(extractor_template % extractor)))

        return extractors

    def on_packet(self, pkt) -> Packet:
        """
        Packet processing callback

        :param pkt: Packet to process
        :type pkt: Packet
        :return: Processed packet
        :rtype: Packet
        """
        extractors = self.build_extractors()
        output = []
        for extractor,processor in extractors:
            try:
                output.append(str(processor(pkt)))
            except Exception as err:
                if self.args.show_exc:
                    sys.stderr.write("-------------------------------------------------------\n")
                    sys.stderr.write(f"Exception raised when evaluating {extractor}:\n")
                    sys.stderr.write(str(err)+"\n")
                    sys.stderr.write("-------------------------------------------------------\n")
                    sys.stderr.flush()
                return pkt

        print(self.args.delimiter.join(output))
        sys.stdout.flush()
        return pkt


    def run(self):
        # Launch pre-run tasks
        self.pre_run()

        # Load any Scapy definition files if provided
        if self.args.loadables is not None:
            for loadable in self.args.loadables:
                l = __import__(loadable)
                for obj in dir(l):
                    o = getattr(l, obj)
                    try:
                        if issubclass(o, Packet) and o != Packet:
                            globals()[obj] = o
                    except TypeError:
                        pass
        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                # Create a Unix socket connector
                connector = UnixConnector(interface)

                # Enable the specified domain in the hub
                ProtocolHub.set_domain(self.args.domain)

                # Set the connector domain
                connector.domain = self.args.domain
                hub = ProtocolHub()
                connector.format = hub.get(self.args.domain).format
                connector.on_packet = self.on_packet

                # Unlock to start processing incoming messages
                connector.unlock()

                connector.join()

        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wextract_main():
    """Launcher for `wextract` CLI application.
    """
    app = WhadExtractApp()
    run_app(app)



from whad.zigbee.stack.apl.zdo.object import ZDOObject
from whad.zigbee.stack.nwk.security import NetworkSecurityMaterial
from whad.zigbee.stack.aps.security import APSNetworkKeyData

from whad.exceptions import RequiredImplementation

from time import sleep
import logging

logger = logging.getLogger(__name__)

class ZDOSecurityManager(ZDOObject):
    """
    ZDO Device Object handling the security-related operations.
    """
    def configure_trust_center(self, trust_center_address):
        """
        Configure the Trust Center address.
        """
        self.zdo.aps_management.set(
            "apsTrustCenterAddress",
            trust_center_address
        )

    def provision_network_key(self, key, key_sequence_number=0):
        """
        Provision a new network key in the NWK layer.
        """
        # Get the Security Material set
        nwkSecurityMaterialSet = self.zdo.nwk_management.get("nwkSecurityMaterialSet")

        # Create a new Network key Material, and add it to the set if it's new
        new_key = NetworkSecurityMaterial(
            key,
            key_sequence_number
        )
        if new_key not in nwkSecurityMaterialSet:
            nwkSecurityMaterialSet.append(new_key)

        # Update the set and configure security level
        self.zdo.nwk_management.set("nwkSecurityMaterialSet", nwkSecurityMaterialSet)
        logger.info("[zdo_security_manager] new network key provisioned.")
        self.zdo.nwk_management.set("nwkSecurityLevel", 5)

    def send_transport_key(self, destination_address):
        """
        Transmit the transport key to a new joining device.
        """
        apsTrustCenterAddress = self.zdo.aps_management.get("apsTrustCenterAddress")
        nwkIeeeAddress = self.zdo.nwk_management.get("nwkIeeeAddress")
        # maybe this test  should be in network manager instead of security manager,
        # since it processes not only transport key but also dev update.
        if apsTrustCenterAddress != 0xFFFFFFFFFFFFFFFF:
            # Centralized network
            if apsTrustCenterAddress == nwkIeeeAddress:
                # We are the trust center, implement 4.6.3.2.2 (p431)
                if self.zdo.configuration.get("allowsJoins"):
                    nwkSecurityMaterialSet = self.zdo.nwk_management.get("nwkSecurityMaterialSet")
                    if len(nwkSecurityMaterialSet) > 0:
                        selected_key = nwkSecurityMaterialSet[0]
                    else:
                        # inject default key if nothing is provided
                        self.provision_network_key(
                            bytes.fromhex("11223344556677881122334455667788"),
                            key_sequence_number = 0
                        )
                        selected_key = nwkSecurityMaterialSet[0]
                    sleep(0.5)
                    self.zdo.aps_management.transport_key(
                        destination_address=destination_address,
                        standard_key_type=1,
                        transport_key_data = APSNetworkKeyData(
                            key=selected_key.key,
                            key_sequence_number=selected_key.key_sequence_number,
                            use_parent=False # assume we are the router
                        )
                    )
                else:
                    # - APSME-REMOVE-DEVICE with:
                    #   + parentAddress = address of the router
                    #   + childAddress = address of the joining non authorized device
                    raise RequiredImplementation("RemoveDeviceFromNetwork")
            else:
                # Here we should trigger:
                # - APSME-UPDATE-DEVICE with:
                #    + destAddr = apsTrustCenterAddress
                #    + deviceAddr = address of the new device
                raise RequiredImplementation("TransportKeyTrustCenterAddress")
        else:
            # Distributed security network
            raise RequiredImplementation("TransportKeyDistributedSecurityNetwork")

    def on_transport_key(self, source_address, standard_key_type, transport_key_data):
        """
        Process all transport keys forwarded by the ZDO.
        """
        # If the key is a network key, provision it and configure the trust center
        if standard_key_type == 1: # process network key
            self.provision_network_key(
                transport_key_data.key,
                transport_key_data.key_sequence_number
            )
            self.configure_trust_center(source_address)

            # Wait until we got an address assigned, then notify authorization to network manager
            while self.zdo.manager.get_layer('nwk').database.get("nwkNetworkAddress") == 0xFFFF:
                pass

            self.zdo.network_manager.on_authorization()



"""WHAD Bluetooth Low Energy base connector

This module provides the :class:`whad.ble.connector.BLE` class that implements
basic BLE-related methods for the attached interface.
"""
import struct
import logging
from typing import Optional

# Scapy
from scapy.layers.bluetooth4LE import BTLE, BTLE_ADV, BTLE_DATA, BTLE_ADV_IND, \
    BTLE_ADV_NONCONN_IND, BTLE_ADV_DIRECT_IND, BTLE_ADV_SCAN_IND, BTLE_SCAN_RSP, \
    BTLE_CTRL
from scapy.packet import Packet

# Device interface
from whad.device import Connector
from whad.hub.discovery import Domain, Capability
from whad.exceptions import UnsupportedDomain, UnsupportedCapability

# Protocol hub
from whad.helpers import message_filter
from whad.hub.generic.cmdresult import Success, CommandResult
from whad.hub.ble.bdaddr import BDAddress
from whad.hub.ble.chanmap import ChannelMap
from whad.hub.ble import Commands, AdvType, Direction, BLEMetadata
from whad.hub.events import ConnectionEvt, DisconnectionEvt, SyncEvt, DesyncEvt, \
    TriggeredEvt, WhadEvent

# Bluetooth Low Energy dependencies
from whad.common.triggers import ManualTrigger, ConnectionEventTrigger, ReceptionTrigger
from whad.ble.profile.advdata import AdvDataFieldList

# Logging
logger = logging.getLogger(__name__)

class BLE(Connector):
    """
    BLE protocol connector.

    This connector drives a BLE-capable device with BLE-specific WHAD messages.
    It is required by various role classes to interact with a real device and pre-process
    domain-specific messages.
    """
    domain = 'ble'

    # correlation table
    SCAPY_CORR_ADV = {
        AdvType.ADV_IND: BTLE_ADV_IND,
        AdvType.ADV_NONCONN_IND: BTLE_ADV_NONCONN_IND,
        AdvType.ADV_DIRECT_IND: BTLE_ADV_DIRECT_IND,
        AdvType.ADV_SCAN_IND: BTLE_ADV_SCAN_IND,
        AdvType.ADV_SCAN_RSP: BTLE_SCAN_RSP
    }

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if isinstance(packet, bytes):
            packet = BTLE(packet)
        return self.hub.ble.format(packet)

    def __init__(self, device=None, synchronous=False):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).

        If `auto` is set to False, PDUs must be processed manually and
        won't be forwarded to PDU-related callbacks.
        """
        self.__started = False
        self.__ready = False
        super().__init__(device)

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Link-layer encryption
        self.__encrypted = False

        # List of active triggers
        self.__triggers = []

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Check device supports BLE
        if not self.device.has_domain(Domain.BtLE):
            raise UnsupportedDomain("Bluetooth Low Energy")

        # Ready
        self.__ready = True

        # Set synchronous mode if provided
        self.enable_synchronous(synchronous)

    @property
    def started(self) -> bool:
        """Check if selected mode is started."""
        return self.__started

    def close(self):
        """Close BLE device
        """
        self.device.close()

    def support_raw_pdu(self) -> bool:
        """
        Determine if the device supports raw PDU.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.BtLE)
            self.__can_send_raw = not capabilities & Capability.NoRawData
        return self.__can_send_raw

    def is_ll_encrypted(self):
        """Determine if the link-layer is encrypted.

        :return: True if link-layer is encrypted, False otherwise
        :rtype: bool
        """

    def can_send(self):
        """
        Determine if the device is able to send PDU
        """
        if self.__can_send is None:
            # Retrieve supported commands
            commands = self.device.get_domain_commands(Domain.BtLE)
            self.__can_send = (
                (commands & (1 << Commands.SendPDU))>0 or
                (commands & (1 << Commands.SendRawPDU))
            )
        return self.__can_send

    def can_scan(self):
        """
        Determine if the device implements a scanner mode.
        """
        # Retrieve supported commands
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.ScanMode))>0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_connect(self):
        """
        Determine if the device can establish a connection as central.
        """
        # Retrieve supported commands
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.ConnectTo))>0

    def can_jam_advertisement_on_channel(self):
        """
        Determine if the device can jam advertisements on a specific channel.
        """
        # Retrieve supported commands
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.JamAdvOnChannel))>0

    def can_be_central(self):
        """
        Determine if the device implements a central mode.
        """
        # Retrieve supported commands
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.CentralMode))>0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_be_peripheral(self):
        """
        Determine if the device implements a peripheral mode.
        """
        # Retrieve supported commands
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.PeripheralMode))>0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_discover_access_addresses(self):
        """
        Determine if the device implements an access addresses discovery mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.SniffAccessAddress)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_sniff_active_connection(self):
        """
        Determine if the device allows to sniff an active connection.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.SniffActiveConn)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_sniff_advertisements(self):
        """
        Determine if the device implements an advertisements sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.SniffAdv)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )


    def can_sniff_new_connection(self):
        """
        Determine if the device implements a new connection sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (
            (commands & (1 << Commands.SniffConnReq)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_inject(self):
        """
        Determine if the device implements an injection mode.
        """
        capabilities = self.device.get_domain_capability(Domain.BtLE)
        return self.can_send() and (capabilities & (1 << Capability.Inject) > 0)


    def can_hijack_master(self):
        """
        Determine if the device implements a master hijacking mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.HijackMaster)) > 0

    def can_hijack_slave(self):
        """
        Determine if the device implements a slave hijacking mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.HijackSlave)) > 0

    def can_hijack_both(self):
        """
        Determine if the device implements a slave and master hijacking mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.HijackBoth)) > 0


    def can_reactive_jam(self):
        """
        Determine if the device implements a reactive jamming mode.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.ReactiveJam)) > 0


    def can_prepare(self):
        """
        Determine if the device can prepare a sequence of packets associated with a trigger.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.PrepareSequence)) > 0

    def can_trigger(self):
        """
        Determine if the device can manually trigger a sequence of packets.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.TriggerSequence)) > 0

    def trigger(self, trigger):
        '''
        Trigger a sequence of packets linked to a Manual Trigger object.
        '''
        if not self.can_trigger():
            raise UnsupportedCapability("Trigger")

        if not isinstance(trigger, ManualTrigger):
            return False

        if trigger.identifier is None:
            return False

        msg = self.hub.ble.create_trigger(trigger.identifier)
        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def can_delete_sequence(self):
        """
        Determine if the device can delete a sequence of packets.
        """
        commands = self.device.get_domain_commands(Domain.BtLE)
        return (commands & (1 << Commands.DeleteSequence)) > 0

    def delete_sequence(self, trigger):
        '''
        Delete a sequence of packets linked to a Trigger object.
        '''
        if not self.can_delete_sequence():
            raise UnsupportedCapability("DeleteSequence")

        if trigger.identifier is None:
            return False

        msg = self.hub.ble.create_delete_sequence(trigger.identifier)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def prepare(self, *packets, trigger=ManualTrigger(), direction=Direction.MASTER_TO_SLAVE):
        """
        Prepare a sequence of packets and associate a trigger to it.
        """
        if not self.can_prepare():
            raise UnsupportedCapability("Prepare")

        prep_packets = []
        for packet in packets:
            if BTLE_DATA in packet:
                prep_packets.append(bytes(packet[BTLE_DATA:]))
            else:
                # Fail if at least one packet has no BTLE_DATA layer
                return False

        if isinstance(trigger, ManualTrigger):
            # Create a prepared sequence with manual trigger
            msg = self.hub.ble.create_prepare_sequence_manual(
                trigger.identifier,
                direction,
                prep_packets
            )
        elif isinstance(trigger, ConnectionEventTrigger):
            # Create a prepared sequence with connection event trigger
            msg = self.hub.ble.create_prepare_sequence_conn_evt(
                trigger.identifier,
                direction,
                trigger.connection_event,
                prep_packets
            )
        elif isinstance(trigger, ReceptionTrigger):
            # Create a prepared sequence with reception trigger
            msg = self.hub.ble.create_prepare_sequence_pattern(
                trigger.identifier,
                direction,
                trigger.pattern,
                trigger.mask,
                trigger.offset,
                prep_packets
            )
        else:
            return False

        resp = self.send_command(msg, message_filter(CommandResult))
        success = isinstance(resp, Success)
        if success:
            self.__triggers.append(trigger)
        return success

    def reactive_jam(self, pattern, position=0, channel=37):
        """
        Performs a reactive jamming attack on provided pattern and channel.
        """
        if not self.can_reactive_jam():
            raise UnsupportedCapability("ReactiveJam")

        # Create a ReactiveJam message
        msg = self.hub.ble.create_reactive_jam(channel, pattern, position)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def jam_advertisement_on_channel(self, channel=37):
        """
        Jam advertisements on a single channel.
        """
        if not self.can_jam_advertisement_on_channel():
            raise UnsupportedCapability("JamAdvOnChannel")

        # Create a JamAdvChan message
        msg = self.hub.ble.create_jam_adv_chan(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def hijack_master(self, access_address):
        """
        Hijack the master role.
        """
        if not self.can_hijack_master():
            raise UnsupportedCapability("Hijack")

        # Create a HijackMaster message
        msg = self.hub.ble.create_hijack_master(access_address)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def discover_access_addresses(self):
        """
        Discover access addresses.
        """
        if not self.can_discover_access_addresses():
            raise UnsupportedCapability("AccessAddressesDiscovery")

        # Create SniffAccessAddress message
        msg = self.hub.ble.create_sniff_access_address(range(37))

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def sniff_active_connection(self, access_address: int , crc_init: int = None,
                                channel_map: ChannelMap = None, hop_interval: int = None,
                                hop_increment: int = None):
        """
        Sniff active connection.
        """
        if not self.can_sniff_active_connection():
            raise UnsupportedCapability("ActiveConnectionSniffing")

        # Create a SniffActiveConn message
        msg = self.hub.ble.create_sniff_active_conn(
            access_address,
            crc_init=crc_init,
            channel_map=channel_map,
            interval=hop_interval,
            increment=hop_increment,
            channels=b"\xFF\xFF\xFF\xFF\x1F"
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def hijack_slave(self, access_address):
        """
        Hijack the slave role.
        """
        if not self.can_hijack_slave():
            raise UnsupportedCapability("Hijack")

        # Create an HijackSlave message
        msg = self.hub.ble.create_hijack_slave(access_address)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def hijack_both(self, access_address):
        """
        Hijack both roles.
        """
        if not self.can_hijack_both():
            raise UnsupportedCapability("Hijack")

        # Create an HijackBoth message
        msg = self.hub.ble.create_hijack_both(access_address)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def sniff_advertisements(self, channel=37, bd_address="FF:FF:FF:FF:FF:FF"):
        """
        Sniff Bluetooth Low Energy advertisements (on a single channel).
        """
        if not self.can_sniff_advertisements():
            raise UnsupportedCapability("Sniff")

        # Create a SniffAdv message
        msg = self.hub.ble.create_sniff_adv(channel, BDAddress(bd_address))

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def sniff_new_connection(self, channel=37, show_advertisements=True,
                             show_empty_packets=False, bd_address="FF:FF:FF:FF:FF:FF"):
        """
        Sniff Bluetooth Low Energy connection (from initiation).
        """
        if not self.can_sniff_new_connection():
            raise UnsupportedCapability("Sniff")

        # Create a SniffConnReq message
        msg = self.hub.ble.create_sniff_connreq(
            channel,
            bd_address=BDAddress(bd_address),
            show_empty=show_empty_packets,
            show_adv=show_advertisements
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_bd_address(self, bd_address, public=True):
        """
        Set Bluetooth Low Energy BD address.
        """
        # Ensure we can spoof BD address
        commands = self.device.get_domain_commands(Domain.BtLE)
        if (commands & (1 << Commands.SetBdAddress))>0:

            # Create a SetBdAddress message
            msg = self.hub.ble.create_set_bd_address(BDAddress(
                bd_address,
                random=(not public)
            ))

            resp = self.send_command(msg, message_filter(CommandResult))
            return isinstance(resp, Success)

        # Cannot set BD address
        return False

    def enable_scan_mode(self, active=False):
        """
        Enable Bluetooth Low Energy scanning mode.
        """
        # Create a ScanMode message
        msg = self.hub.ble.create_scan_mode(active=active)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def enable_central_mode(self):
        """
        Enable Bluetooth Low Energy central mode (acts as master).
        """
        # Create a CentalMode message
        msg = self.hub.ble.create_central_mode()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def enable_adv_mode(self, adv_data=None, scan_data=None):
        """
        Enable BLE advertising mode (acts as a broadcaster)
        """
        # Create a AdvMode message
        msg = self.hub.ble.create_adv_mode(
            adv_data,
            scan_rsp=scan_data
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def enable_peripheral_mode(self, adv_data: bytes = None, scan_data: bytes = None):
        """
        Enable Bluetooth Low Energy peripheral mode (acts as slave).
        """
        # Build advertising data if required
        if isinstance(adv_data, AdvDataFieldList):
            adv_data = adv_data.to_bytes()
        if isinstance(scan_data, AdvDataFieldList):
            scan_data = scan_data.to_bytes()

        # Create a PeriphMode message
        msg = self.hub.ble.create_periph_mode(
            adv_data=adv_data,
            scan_rsp=scan_data
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def connect_to(self, bd_addr: BDAddress, random: Optional[bool] = False, access_address: Optional[int] = None,
                   channel_map: Optional[ChannelMap] = None, crc_init: Optional[int] = None, hop_interval: Optional[int] = None,
                   hop_increment: Optional[int] = None):
        """
        Initiate a Bluetooth Low Energy connection.
        """
        # Create a ConnectTo message
        msg = self.hub.ble.create_connect_to(
            bd_address=BDAddress(bd_addr, random=random),
            access_address=access_address,
            channel_map=channel_map,
            interval=hop_interval,
            increment=hop_increment,
            crc_init=crc_init
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def start(self):
        """
        Start currently enabled mode.
        """
        if not self.__started:
            logger.info('starting current BLE mode ...')
            # Create and send a Start message
            msg = self.hub.ble.create_start()
            resp = self.send_command(msg, message_filter(CommandResult))

            # Check result
            if isinstance(resp, Success):
                logger.info('current BLE mode successfully started')
                self.__started = True
            else:
                logger.info('an error occurred while starting !')
                self.__started = False

            return self.__started
        else:
            logger.debug("starting current BLE mode, ignoring (already started)")
            return True

    def disconnect(self, conn_handle):
        """Terminate a specific connection.

        :param int conn_handle: Connection handle of the connection to terminate.
        """
        logger.debug("[ble] sending Disconnect message")
        # Create a Disconnect message
        msg = self.hub.ble.create_disconnect(conn_handle)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def stop(self):
        """
        Stop currently enabled mode.
        """
        if self.__started:
            logger.debug("stopping BLE connector ...")

            # Create and send a Stop message
            msg = self.hub.ble.create_stop()
            resp = self.send_command(msg, message_filter(CommandResult))

            # Successfully stopped ?
            if isinstance(resp, Success):
                # Remove all triggers
                for trigger in self.__triggers:
                    trigger.connector = None
                self.__triggers = []

                # Connector is now stopped.
                self.__started = False

                logger.debug("BLE connector is now stopped.")
                return True

            # Return result.
            return False
        else:
            logger.debug("Stopping current BLE connector, ignored (already stopped)")
            return True

    def set_encryption(self, conn_handle, enabled=False, ll_key=None, ll_iv=None,
                       key=None, rand=None, ediv=None):
        """Notify WHAD device about encryption status
        """

        #print("set_encryption", enabled, ll_key.hex(), ll_iv.hex(), key.hex(), rand, ediv)

        # Create a SetEncryption message
        msg = self.hub.ble.create_set_encryption(
            conn_handle,
            ll_key if ll_key is not None else b"",
            ll_iv if ll_iv is not None else b"",
            key if key is not None else b"",
            struct.pack('<Q', rand) if rand is not None else b"",
            struct.pack('<H', ediv) if ediv is not None else b"",
            enabled
        )

        resp = self.send_command(msg, message_filter(CommandResult))

        # Update LL encryption status
        if isinstance(resp, Success):
            self.__encrypted = enabled
            logger.info('[ble connector] encryption is now: %s', self.__encrypted)

        # Return command result
        return isinstance(resp, Success)

    def on_generic_msg(self, message):
        """Generic message callback.
        """
        logger.info('generic message: %s', message)

    def on_discovery_msg(self, message):
        """Discovery message callback.
        """
        logger.info('discovery message: %s', message)

    def on_domain_msg(self, domain, message):
        """Domain-related message callback.
        """
        return

    def on_synchronized(self, access_address=None, crc_init=None, hop_increment=None,
                        hop_interval=None, channel_map=None):
        """Connection synchronization event callback.
        """
        return

    def on_desynchronized(self, access_address=None):
        """Connection desynchronization callback.
        """
        return

    def on_event(self, event: WhadEvent):
        """WHAD event callback handler.

        :param event: WHAD event to process
        :type event: :class:`whad.hub.events.WhadEvent`
        """

        # Don't process if connector is not ready
        if not self.__ready:
            logger.debug("[ble connector] on_event() called, but connector is not ready")
            return

        # Dispatch event
        if isinstance(event, ConnectionEvt):
            logger.debug("[ble connector] on_event() called: connection event")
            self.on_connected(event)
        elif isinstance(event, DisconnectionEvt):
            logger.debug("[ble connector] on_event() called: disconnection event")
            self.on_disconnected(event)
        elif isinstance(event, SyncEvt):
            logger.debug("[ble connector] on_event() called: sync event")
            self.on_synchronized(
                access_address = event.access_address,
                crc_init = event.crc_init,
                hop_interval = event.hop_interval,
                hop_increment = event.hop_increment,
                channel_map = event.channel_map
            )
        elif isinstance(event, DesyncEvt):
            logger.debug("[ble connector] on_event() called: desync event")
            self.on_desynchronized(access_address=event.access_address)
        elif isinstance(event, TriggeredEvt):
            logger.debug("[ble connector] on_event() called: triggered event")
            self.on_triggered(event.id)
        else:
            logger.error("[ble connector] on_event() called: unknown event !")

    def on_packet(self, packet: Packet):
        """Dispatch incoming packet.

        :param packet: Incoming packet
        :type packet: :class:`scapy.packet.Packet`
        """
        logger.debug('[BLE connector] on_packet')
        # discard processed packets or if we're not ready
        if hasattr(packet, "processed") and packet.metadata.processed or not self.__ready:
            return

        if BTLE_ADV in packet:
            adv_pdu = packet[BTLE_ADV:]
            adv_pdu.metadata = packet.metadata
            self.on_adv_pdu(adv_pdu)
        elif BTLE_DATA in packet:
            conn_pdu = packet[BTLE_DATA:]
            conn_pdu.metadata = packet.metadata
            if packet.LLID == 3:
                self.on_ctl_pdu(conn_pdu)
            elif packet.LLID in (1,2):
                self.on_data_pdu(conn_pdu)
            else:
                self.on_error_pdu(conn_pdu)

    def on_adv_pdu(self, packet: Packet):
        """Advertisement PDU callback
        """
        logger.info("received an advertisement PDU")

    def on_connected(self, connection_data: dict):
        """Connection event callback
        """
        logger.info("a connection has been established")
        logger.debug(
            f"connection handle: {connection_data.conn_handle}" if connection_data.conn_handle is not None else 0
        )

    def on_triggered(self, identifier):
        """Prepare sequence triggered event callback
        """
        for trigger in self.__triggers:
            if trigger.identifier == identifier:
                trigger.triggered = True

    def on_disconnected(self, disconnection_data: dict):
        """Disconnection event handler.
        """
        logger.info("a connection has been terminated")
        for trigger in self.__triggers:
            self.delete_sequence(trigger)

    def on_data_pdu(self, pdu):
        """Data PDU handler.
        """
        logger.info("received a data PDU")

    def on_ctl_pdu(self, pdu):
        """Control PDU handler.
        """
        logger.info("received a control PDU")

    def on_error_pdu(self, pdu):
        """Error PDU handler
        """
        return

    def send_ctrl_pdu(self, pdu, conn_handle=0, direction=Direction.MASTER_TO_SLAVE,
                      access_address=0x8e89bed6, encrypt=None):
        """Send control PDU (requires raw injection capability).
        """
        logger.info("send control PDU to connection (handle:%d, direction: %d)", conn_handle, direction)
        return self.send_pdu(pdu, conn_handle=conn_handle, direction=direction,
                             access_address=access_address, encrypt=encrypt)

    def send_data_pdu(self, data, conn_handle=0, direction=Direction.MASTER_TO_SLAVE,
                      access_address=0x8e89bed6, encrypt=None):
        """Send data (L2CAP) PDU.
        """
        logger.info("send data PDU to connection (handle:%d, direction: %d)", conn_handle, direction)
        return self.send_pdu(data, conn_handle=conn_handle, direction=direction,
                             access_address=access_address, encrypt=encrypt)

    def send_pdu(self, pdu, conn_handle=0, direction=Direction.MASTER_TO_SLAVE,
                 access_address=0x8e89bed6, encrypt=None):
        """Send a generic BLE PDU.
        """
        if self.can_send():
            if self.support_raw_pdu():
                packet = BTLE(access_addr=access_address)/pdu
                send_raw = True
            else:
                # Sanity check (issue #183): if there is a BTLE_CTRL layer in this
                # PDU and we don't support raw PDU, issue an error and raise
                # an UnsupportedCapability exception.
                #
                # If we are using an HCI adapter, control PDUs are allowed because
                # we need this to handle pairing and bonding.
                if BTLE_CTRL in pdu and not self.device.interface.startswith("hci"):
                    logger.error((
                        "WHAD interface %s cannot send BLE control PDUs, please "
                        "use another interface that supports sending raw PDUs."
                    ), self.device.interface)
                    raise UnsupportedCapability("RawInject")

                # PDU is not raw
                packet = pdu
                send_raw = False

            # Build packet metadata in order to get this packet correctly
            # translated into the corresponding WHAD message.
            packet.metadata = BLEMetadata()
            packet.metadata.direction = direction
            packet.metadata.connection_handle = conn_handle
            packet.metadata.raw = send_raw

            # If encrypt is provided, take it into account
            # otherwise consider using the internal link-layer encryption status
            if encrypt is not None and isinstance(encrypt, bool):
                logger.info("[ble connector] encrypt is specified (%s)", encrypt)
                packet.metadata.encrypt = True
            else:
                logger.info("[ble connector] link-layer encryption: %s", self.__encrypted)
                packet.metadata.encrypt = False

            # Send BLE packet
            return super().send_packet(packet)

        # Cannot send BLE packet
        return False

    def send_packet(self, packet: Packet):
        """Packet send hook

        This hook makes sure we are using a valid WHAD message when sending
        a packet, if this method is called from outside.

        :param packet: Packet to send
        :type packet: :class:`scapy.packet.Packet`
        :return: True if packet has correctly been sent, False otherwise.
        """

        if self.support_raw_pdu():
            # We expect a BTLE header for raw packets
            if BTLE not in packet and BTLE_DATA in packet:
                # Add a BTLE layer, copy metadata and mark packet
                # as raw.
                packet_ = BTLE(access_addr=0x11223344) / packet[BTLE_DATA]
                packet_.metadata = packet.metadata
                packet_.metadata.raw = True
        else:
            # We don't expect BTLE headers for non-raw packets
            if BTLE in packet:
                # Extract the BTLE_DATA layer, copy metadata and mark
                # packet as non-raw.
                packet_ = packet[BTLE_DATA]
                packet_.metadata = packet.metadata
                packet_.metadata.raw = False

        # Send BLE packet
        return super().send_packet(packet)

    def on_mtu_changed(self, conn_handle, mtu: int):
        """Connection MTU has been updated.
        """




from whad.common.metadata import Metadata
from dataclasses import dataclass
from whad.scapy.layers.esb import ESB_Hdr
from whad.esb.esbaddr import ESBAddress

from whad.hub.unifying import UnifyingMetadata, generate_unifying_metadata



from scapy.packet import Packet, bind_layers, split_layers
from scapy.fields import ByteEnumField, StrLenField, LEIntField, XShortField, LEShortField, \
    FieldLenField, StrFixedLenField, ConditionalField, PacketField, XLongField, XIntField,  XLEIntField, \
    XLEShortField, BitEnumField, BitField, ByteField, FieldListField, StrField, XByteField
from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS, Dot15d4Data
from scapy.config import conf

RC_COMMAND_CODES = {
 0x40 : "TV POWER",
 0x43 : "MUTE",
 0x6C : "ALL POWER Off:",
 0x6D : "On:",
 0x41 : "VOL+",
 0x42 : "VOL–",
 0x4C : "REPLAY",
 0x30 : "CH+",
 0x31 : "CH–",
 0x48 : "REWIND",
 0x61 : "PLAY/PAUSE",
 0x49 : "FAST FORWARD",
 0x0D : "EXIT",
 0x09 : "MENU",
 0x47 : "RECORD",
 0x53 : "GUIDE",
 0x37 : "PAGE UP",
 0x01 : "UP ARROW",
 0x03 : "LEFT ARROW",
 0x00 : "OK",
 0x04 : "RIGHT ARROW",
 0x02 : "DOWN ARROW",
 0x32 : "LAST",
 0x35 : "INFO",
 0x38 : "PAGE DOWN",
 0x74 : "OCAP A (yellow triangle) - Function Key 3",
 0x71 : "OCAP B (blue square) - Function Key 2",
 0x72 : "OCAP C (red circle) - Function Key 0",
 0x73 : "OCAP D (green diamond) - Function Key 1",
 0x20 : "0",
 0x21 : "1",
 0x22 : "2",
 0x23 : "3",
 0x24 : "4",
 0x25 : "5",
 0x26 : "6",
 0x27 : "7",
 0x28 : "8",
 0x29 : "9",
 0x4B : "30-Second Skip Ahead",
 0x34 : "Input/Select",
 0x10 : "HOME",
 0xA0 : "PROFILE",
 0xA1 : "CALL",
 0xA2 : "HOLD",
 0xA3 : "END",
 0xA4 : "VIEWS",
 0xA5 : "SELF-VIEW",
 0xA6 : "ZOOM IN",
 0xA7 : "ZOOM OUT",
 0x65 : "MUTE MIC",
 0x64 : "STOP VIDEO",
 0xB0 : "A",
 0xB1 : "B",
 0xB2 : "C",
 0xB3 : "D",
 0xB4 : "E",
 0xB5 : "F",
 0xB6 : "G",
 0xB7 : "H",
 0xB8 : "I",
 0xB9 : "J",
 0xBA : "K",
 0xBB : "L",
 0xBC : "M",
 0xBD : "N",
 0xBE : "O",
 0xBF : "P",
 0xC0 : "Q",
 0xC1 : "R",
 0xC2 : "S",
 0xC3 : "T",
 0xC4 : "U",
 0xC5 : "V",
 0xC6 : "W",
 0xC7 : "X",
 0xC8 : "Y",
 0xC9 : "Z",
 0xCA : "a",
 0xCB : "b",
 0xCC : "c",
 0xCD : "d",
 0xCE : "e",
 0xCF : "f",
 0xD0 : "g",
 0xD1 : "h",
 0xD2 : "i",
 0xD3 : "j",
 0xD4 : "k",
 0xD5 : "l",
 0xD6 : "m",
 0xD7 : "n",
 0xD8 : "o",
 0xD9 : "p",
 0xDA : "q",
 0xDB : "r",
 0xDC : "s",
 0xDD : "t",
 0xDE : "u",
 0xDF : "v",
 0xE0 : "w",
 0xE1 : "x",
 0xE2 : "y",
 0xE3 : "z",
 0xE4 : "?",
 0xE5 : "!",
 0xE6 : "#",
 0xE7 : "$",
 0xE8 : "%",
 0xE9 : "&",
 0xEA : "*",
 0xEB : "(",
 0xEC : ")",
 0xED : "+",
 0xEE : "-",
 0xEF : "=",
 0xF0 : "/",
 0xF1 : "_",
 0xF2 : "\"",
 0xF3 : ":",
 0xF4 : ";",
 0xF5 : "@",
 0x2A : ".",
 0xF6 : "'",
 0xF7 :",",
 0xA8 : "BACKSPACE",
 0x2B : "RETURN",
 0xA9 : "LOCK/UNLOCK",
 0xAA : "CAPS",
 0xAB : "ALT",
 0xAC : "SPACE",
 0xAD : "www.",
 0xAE : ".com",
 0xF8 : "On Demand",
 0xF9 : "RF Bypass",
 0xFA : "Next Favorite Channel",
}

class RF4CE_Hdr(Packet):
    name = "RF4CE Packet"
    fields_desc = [
        BitField("channel_identifier", 0, 2),
        BitField("reserved", 0, 1),
        BitField("protocol_version", 0, 2),
        BitEnumField("security_enabled", 0, 1, {0: "disabled", 1:"enabled"}),
        BitEnumField("frame_type", 0, 2, {0: "reserved", 1:"data", 2:"command", 3:"vendor"}),

        ConditionalField(
            XLEIntField("mic", None),
            lambda pkt: pkt.security_enabled  == 1
        ),

        LEIntField("frame_counter", None),
    ]

    def pre_dissect(self,s):
        if ((s[0] & 0b100) >> 2) == 1:
            return s[:1] + s[-4:] + s[1:-4]
        else:
            return s

    def post_build(self,p,pay):
        if self.security_enabled == 1:
            mic = self.mic
            return p[:1] + p[5:] + pay + p[1:5]
        else:
            return p + pay

    def post_dissect(self, s):
        """Override layer post_dissect() function to reset raw packet cache.
        """
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s
        
class RF4CE_Vendor_Hdr(Packet):
    name = "RF4CE Vendor Data Packet"
    fields_desc = [
        XByteField("profile_id", None),
        LEShortField("vendor_id", None),
    ]

class RF4CE_Data_Hdr(Packet):
    name = "RF4CE Data Packet"
    fields_desc = [
        XByteField("profile_id", None),
        LEShortField("vendor_id", None),
    ]

class RF4CE_Command_Hdr(Packet):
    name = "RF4CE Command Packet"
    fields_desc = [
        ByteEnumField("command_identifier", None, {
            0x01 : "discovery_req",
            0x02 : "discovery_rsp",
            0x03 : "pair_req",
            0x04 : "pair_rsp",
            0x05 : "unpair_req",
            0x06 : "key_seed",
            0x07 : "ping_req",
            0x08 : "ping_rsp"
        })
    ]

class RF4CE_Cmd_Discovery_Request(Packet):
    name = "RF4CE Discovery Request Command"
    fields_desc = [
        BitField("reserved2", 0, 4),
        BitEnumField("channel_normalization_capable", 0, 1, {0 : "no", 1 : "yes"}),
        BitEnumField("security_capable", 0,  1,{0 : "no", 1 : "yes"}),
        BitEnumField("power_source", 0,  1,{ 0 : "battery_source", 1 : "alternating_current_source"}),
        BitEnumField("node_type", 0,  1,{0 : "controller", 1 : "target"}),

        LEShortField("vendor_identifier",0),
        StrFixedLenField("vendor_string", b"", length=7),

        BitField("reserved4", 0, 1),
        BitField("number_of_supported_profiles", 0, 3),
        BitField("reserved3", 0, 1),
        BitField("number_of_supported_device_types", 0, 2),
        BitEnumField("user_string_specificied", 0,  1, {0 : "no", 1 : "yes"}),
        ConditionalField(
            StrFixedLenField("user_string", None, length=15),
            lambda pkt: pkt.user_string_specificied  == 1
        ),
        FieldListField("device_type_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_device_types),
        FieldListField("profile_identifier_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_profiles),
        ByteField("requested_device_type", None)
    ]

class RF4CE_Cmd_Discovery_Response(Packet):
    name = "RF4CE Discovery Response Command"
    fields_desc = [
        ByteField("status", None),

        BitField("reserved2", 0, 4),
        BitEnumField("channel_normalization_capable", 0, 1, {0 : "no", 1 : "yes"}),
        BitEnumField("security_capable", 0,  1,{0 : "no", 1 : "yes"}),
        BitEnumField("power_source", 0,  1,{ 0 : "battery_source", 1 : "alternating_current_source"}),
        BitEnumField("node_type", 0,  1,{0 : "controller", 1 : "target"}),


        LEShortField("vendor_identifier",0),
        StrFixedLenField("vendor_string", b"", length=7),

        BitField("reserved4", 0, 1),
        BitField("number_of_supported_profiles", 0, 3),
        BitField("reserved3", 0, 1),
        BitField("number_of_supported_device_types", 0, 2),
        BitEnumField("user_string_specificied", 0,  1, {0 : "no", 1 : "yes"}),
        ConditionalField(
            StrFixedLenField("user_string", None, length=15),
            lambda pkt: pkt.user_string_specificied  == 1
        ),
        FieldListField("device_type_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_device_types),
        FieldListField("profile_identifier_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_profiles),

        ByteField("discovery_req_lqi", None)
    ]


class RF4CE_Cmd_Pair_Request(Packet):
    name = "RF4CE Pair Request Command"
    fields_desc = [
        XLEShortField("nwk_addr", None),

        BitField("reserved2", 0, 4),
        BitEnumField("channel_normalization_capable", 0, 1, {0 : "no", 1 : "yes"}),
        BitEnumField("security_capable", 0,  1,{0 : "no", 1 : "yes"}),
        BitEnumField("power_source", 0,  1,{ 0 : "battery_source", 1 : "alternating_current_source"}),
        BitEnumField("node_type", 0,  1,{0 : "controller", 1 : "target"}),


        LEShortField("vendor_identifier",None),
        StrFixedLenField("vendor_string", b"", length=7),

        BitField("reserved4", 0, 1),
        BitField("number_of_supported_profiles", 0, 3),
        BitField("reserved3", 0, 1),
        BitField("number_of_supported_device_types", 0, 2),
        BitEnumField("user_string_specificied", 0,  1, {0 : "no", 1 : "yes"}),
        ConditionalField(
            StrFixedLenField("user_string", None, length=15),
            lambda pkt: pkt.user_string_specificied  == 1
        ),
        FieldListField("device_type_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_device_types),
        FieldListField("profile_identifier_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_profiles),

        ByteField("key_exchange_transfer_count", None)
    ]



class RF4CE_Cmd_Pair_Response(Packet):
    name = "RF4CE Pair Response Command"
    fields_desc = [
        ByteField("status", None),
        XLEShortField("allocated_nwk_addr", None),
        XLEShortField("nwk_addr", None),

        BitField("reserved2", 0, 4),
        BitEnumField("channel_normalization_capable", 0, 1, {0 : "no", 1 : "yes"}),
        BitEnumField("security_capable", 0,  1,{0 : "no", 1 : "yes"}),
        BitEnumField("power_source", 0,  1,{ 0 : "battery_source", 1 : "alternating_current_source"}),
        BitEnumField("node_type", 0,  1,{0 : "controller", 1 : "target"}),


        LEShortField("vendor_identifier",None),
        StrFixedLenField("vendor_string", b"", length=7),

        BitField("reserved4", 0, 1),
        BitField("number_of_supported_profiles", 0, 3),
        BitField("reserved3", 0, 1),
        BitField("number_of_supported_device_types", 0, 2),
        BitEnumField("user_string_specificied", 0,  1, {0 : "no", 1 : "yes"}),
        ConditionalField(
            StrFixedLenField("user_string", None, length=15),
            lambda pkt: pkt.user_string_specificied  == 1
        ),
        FieldListField("device_type_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_device_types),
        FieldListField("profile_identifier_list", None,
            ByteField("type_list", None)
        , count_from=lambda pkt:pkt.number_of_supported_profiles),

    ]

class RF4CE_Cmd_Unpair_Request(Packet):
    name = "RF4CE Unpair Request Command"
    fields_desc = []


class RF4CE_Cmd_Key_Seed(Packet):
    name = "RF4CE Key Seed Command"
    fields_desc = [
        ByteField("key_sequence_number", None),
        StrFixedLenField("seed_data", None, length=80),
    ]

class RF4CE_Cmd_Ping_Request(Packet):
    name = "RF4CE Ping Request Command"
    fields_desc = [
        ByteField("ping_options", None),
        StrField("ping_payload", None),
    ]


class RF4CE_Cmd_Ping_Response(Packet):
    name = "RF4CE Ping Response Command"
    fields_desc = [
        ByteField("ping_options", None),
        StrField("ping_payload", None),
    ]


class RF4CE_Vendor_ZRC_Hdr(Packet):
    name = "RF4CE Vendor Data - ZRC Profile"
    fields_desc = [
        BitField("reserved1", 0, 3),
        BitEnumField("command_code", 0, 5, {
            0x01 : "user_control_pressed",
            0x02 : "user_control_repeated",
            0x03 : "user_control_released",
            0x04 : "command discovery request",
            0x05 : "command discovery response"
        }),
    ]


class RF4CE_Vendor_ZRC_User_Control_Pressed(Packet):
    name = "RF4CE Vendor Data - User Control Pressed - ZRC Profile"
    fields_desc = [
        ByteEnumField("code", None, RC_COMMAND_CODES),
        StrField("ctrl_payload", None)
    ]

class RF4CE_Vendor_ZRC_User_Control_Released(Packet):
    name = "RF4CE Vendor Data - User Control Released - ZRC Profile"
    fields_desc = []

class RF4CE_Vendor_ZRC_User_Control_Repeated(Packet):
    name = "RF4CE Vendor Data - User Control Repeated - ZRC Profile"
    fields_desc = []

class RF4CE_Vendor_ZRC_Command_Discovery_Request(Packet):
    name = "RF4CE Vendor Data - Command Discovery Request - ZRC Profile"
    fields_desc = [
        ByteField("command_bank_number", None),
    ]

class RF4CE_Vendor_ZRC_Command_Discovery_Response(Packet):
    name = "RF4CE Vendor Data - Command Discovery Response - ZRC Profile"
    fields_desc = [
        ByteField("command_bank_number", None),
        StrFixedLenField("commands_supported", b"", length=256),
    ]

class RF4CE_Vendor_MSO_Hdr(Packet):
    name = "RF4CE Vendor Data - MSO Profile"
    fields_desc = [
        ByteEnumField("command_code", None, {
            0x01 : "user_control_pressed",
            0x02 : "user_control_repeated",
            0x03 : "user_control_released",
            0x20 : "check_validation_request",
            0x21 : "check_validation_response",
            0x22 : "set_attribute_request",
            0x23 : "set_attribute_response",
            0x24 : "get_attribute_request",
            0x25 : "get_attribute_response",
            0x32 : "audio"
        }),
    ]

class RF4CE_Vendor_MSO_Audio(Packet):
    name = "RF4CE Vendor Data - Audio - MSO Profile (Telink extension)"
    fields_desc = [
        ByteField("sequence_number", None),
        ByteEnumField("audio_cmd_id", None, {
            0x01 : "start_req",
            0x02 : "stop_req",
            0x03 : "data_notify",
            0x81 : "start_rsp",
            0x82 : "stop_rsp",
            0x83 : "data_rsp" # not implemented ?
        }),
        ByteField("length", 0),
    ]

    def post_build(self, p, pay):
        if self.length is None:
            self.length = len(pay)
        return p + pay

class RF4CE_Vendor_MSO_Audio_Start_Request(Packet):
    name = "RF4CE Vendor Data - Audio Start Request"
    fields_desc = [
        LEShortField("sample_rate", None),
        ByteField("resolution_bits", None),
        ByteField("mic_channel_number", None),
        ByteEnumField("codec_type", None, {1 : "ADPCM"}),
        ByteField("packet_size", None),
        ByteField("interval", None),
        ByteField("channel_number", None),
        ByteField("duration", None),

    ]

class RF4CE_Vendor_MSO_Audio_Stop_Request(Packet):
    name = "RF4CE Vendor Data - Audio Stop Request"
    fields_desc = []


class RF4CE_Vendor_MSO_Audio_Start_Response(Packet):
    name = "RF4CE Vendor Data - Audio Start Response"
    fields_desc = [
        ByteField("best_channel", None)
    ]


class RF4CE_Vendor_MSO_Audio_Stop_Response(Packet):
    name = "RF4CE Vendor Data - Audio Stop Response"
    fields_desc = [
        ByteField("state", None)
    ]


class RF4CE_Vendor_MSO_Audio_Data_Notify(Packet):
    name = "RF4CE Vendor Data - Audio Data Notify"
    fields_desc = [
        LEIntField("header", None),
        StrFixedLenField("samples", None, length=80)
    ]


class RF4CE_Vendor_MSO_User_Control_Pressed(Packet):
    name = "RF4CE Vendor Data - User Control Pressed - MSO Profile"
    fields_desc = [
        ByteEnumField("code", None, RC_COMMAND_CODES),
        StrField("ctrl_payload", None)
    ]

class RF4CE_Vendor_MSO_User_Control_Released(Packet):
    name = "RF4CE Vendor Data - User Control Released - MSO Profile"
    fields_desc = []

class RF4CE_Vendor_MSO_User_Control_Repeated(Packet):
    name = "RF4CE Vendor Data - User Control Repeated - MSO Profile"
    fields_desc = []

class RF4CE_Vendor_MSO_Check_Validation_Request(Packet):
    name = "RF4CE Vendor Data - Check Validation Request - MSO Profile"
    fields_desc = [
        BitField("reserved1", 0, 7),
        BitEnumField("request_automatic_validation", 0, 1, {0:"no", 1:"yes"})
    ]


class RF4CE_Vendor_MSO_Check_Validation_Response(Packet):
    name = "RF4CE Vendor Data - Check Validation Response - MSO Profile"
    fields_desc = [
        ByteEnumField("check_validation_status", None, {
            0x0 : "success",
            0xc0 : "pending",
            0xc1 : "timeout",
            0xc2 : "collision",
            0xc3 : "failure",
            0xc4 : "abort",
            0xc5 : "full_abort",
        })
    ]

MSO_REMOTE_INFORMATION_BASE = {
    0x00 : "Peripheral IDs",
    0x01 : "RF Statistics",
    0x02 : "Versioning",
    0x03 : "Battery Status",
    0x04 : "Short RF Retry Period",
    0xDB : "IR-RF Database",
    0xDC : "Validation Configuration",
    0xFF : "General Purpose",
}

class RF4CE_Vendor_MSO_Set_Attribute_Request(Packet):
    name = "RF4CE Vendor Data - Set Attribute Request - MSO Profile"
    fields_desc = [
        ByteEnumField("attribute_identifier", None, MSO_REMOTE_INFORMATION_BASE),
        ByteField("index", None),
        FieldLenField("value_length", None, length_of="value", fmt="B"),
        StrLenField("value", None, length_from=lambda x:x.value_length)
    ]

class RF4CE_Vendor_MSO_Set_Attribute_Response(Packet):
    name = "RF4CE Vendor Data - Set Attribute Response - MSO Profile"
    fields_desc = [
        ByteEnumField("attribute_identifier", None, MSO_REMOTE_INFORMATION_BASE),
        ByteField("index", None),
        ByteField("status", None)
    ]

class RF4CE_Vendor_MSO_Get_Attribute_Request(Packet):
    name = "RF4CE Vendor Data - Get Attribute Request - MSO Profile"
    fields_desc = [
        ByteEnumField("attribute_identifier", None, MSO_REMOTE_INFORMATION_BASE),
        ByteField("index", None),
        ByteField("value_length", None)
    ]

class RF4CE_Vendor_MSO_Get_Attribute_Response(Packet):
    name = "RF4CE Vendor Data - Get Attribute Response - MSO Profile"
    fields_desc = [
        ByteEnumField("attribute_identifier", None, MSO_REMOTE_INFORMATION_BASE),
        ByteField("index", None),
        ByteField("status", None),
        FieldLenField("value_length", None, length_of="value", fmt="B"),
        StrLenField("value", None, length_from=lambda x:x.value_length)
    ]

bind_layers(RF4CE_Hdr, RF4CE_Command_Hdr, frame_type = 2)
bind_layers(RF4CE_Hdr, RF4CE_Data_Hdr, frame_type = 1)
bind_layers(RF4CE_Hdr, RF4CE_Vendor_Hdr, frame_type = 3)

bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Discovery_Request, command_identifier = 1)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Discovery_Response, command_identifier = 2)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Pair_Request, command_identifier = 3)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Pair_Response, command_identifier = 4)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Unpair_Request, command_identifier = 5)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Key_Seed, command_identifier = 6)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Ping_Request, command_identifier = 7)
bind_layers(RF4CE_Command_Hdr, RF4CE_Cmd_Ping_Response, command_identifier = 8)

bind_layers(RF4CE_Vendor_Hdr, RF4CE_Vendor_MSO_Hdr, profile_id = 0x01)


bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_User_Control_Pressed, command_code = 0x01)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_User_Control_Repeated, command_code = 0x02)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_User_Control_Released, command_code = 0x03)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Check_Validation_Request, command_code = 0x20)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Check_Validation_Response, command_code = 0x21)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Set_Attribute_Request, command_code = 0x22)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Set_Attribute_Response, command_code = 0x23)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Get_Attribute_Request, command_code = 0x24)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Get_Attribute_Response, command_code = 0x25)
bind_layers(RF4CE_Vendor_MSO_Hdr, RF4CE_Vendor_MSO_Audio, command_code = 0x32)

bind_layers(RF4CE_Vendor_Hdr, RF4CE_Vendor_MSO_Hdr,  profile_id = 0xc0)

bind_layers(RF4CE_Vendor_ZRC_Hdr, RF4CE_Vendor_ZRC_User_Control_Pressed, command_code = 0x01)
bind_layers(RF4CE_Vendor_ZRC_Hdr, RF4CE_Vendor_ZRC_User_Control_Repeated, command_code = 0x02)
bind_layers(RF4CE_Vendor_ZRC_Hdr, RF4CE_Vendor_ZRC_User_Control_Released, command_code = 0x03)
bind_layers(RF4CE_Vendor_ZRC_Hdr, RF4CE_Vendor_ZRC_Command_Discovery_Request, command_code = 0x04)
bind_layers(RF4CE_Vendor_ZRC_Hdr, RF4CE_Vendor_ZRC_Command_Discovery_Response, command_code = 0x05)


bind_layers(RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Start_Request, audio_cmd_id = 0x01)
bind_layers(RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Stop_Request, audio_cmd_id = 0x02)
bind_layers(RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Data_Notify, audio_cmd_id = 0x03)

bind_layers(RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Start_Response, audio_cmd_id = 0x81)
bind_layers(RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Stop_Response, audio_cmd_id = 0x82)

# Monkey patch to add RF4CE support in Dot15d4 layer
old_guess_payload_class = Dot15d4Data.guess_payload_class

def new_guess_payload_class(self, payload):
    if conf.dot15d4_protocol == "rf4ce":
        return RF4CE_Hdr
    else:
        return old_guess_payload_class(self, payload)

Dot15d4Data.guess_payload_class = new_guess_payload_class



from whad.dot15d4.stack.mac.constants import MACDeviceType, MACPowerSource
from whad.zigbee.stack.apl.constants import LogicalDeviceType

class NodeDescriptor:
    """
    This class is a descriptor for a ZigBee Node.
    """
    def __init__(
                    self,
                    logical_type=LogicalDeviceType.END_DEVICE,
                    complex_descriptor_available=False,
                    user_descriptor_available=False,
                    aps_flags=0,
                    support_868_mhz=False,
                    support_902_mhz=False,
                    support_2400_mhz=True,
                    alternate_pan_coordinator=False,
                    device_type=MACDeviceType.FFD,
                    power_source=MACPowerSource.ALTERNATING_CURRENT_SOURCE,
                    receiver_on_when_idle=True,
                    security_capability=True,
                    allocate_address=False,
                    manufacturer_code=0x1234,
                    max_buffer_size=128,
                    max_incoming_transfer_size=128,
                    server_primary_trust_center = False,
                    server_backup_trust_center = False,
                    server_primary_binding_table_cache = False,
                    server_backup_binding_table_cache = False,
                    server_primary_discovery_cache = False,
                    server_backup_discovery_cache = False,
                    network_manager = False,
                    stack_compliance_revision = 21,
                    max_outgoing_transfer_size = 128,
                    extended_active_endpoint_list_available = False,
                    extended_simple_descriptors_list_available = False
        ):
        self.logical_type = logical_type
        self.complex_descriptor_available = complex_descriptor_available
        self.user_descriptor_available = user_descriptor_available
        self.aps_flags = aps_flags
        self.support_868_mhz = support_868_mhz
        self.support_902_mhz = support_902_mhz
        self.support_2400_mhz = support_2400_mhz
        self.alternate_pan_coordinator = alternate_pan_coordinator
        self.device_type = device_type
        self.power_source = power_source
        self.receiver_on_when_idle = receiver_on_when_idle
        self.security_capability = security_capability
        self.allocate_address = allocate_address
        self.manufacturer_code = manufacturer_code
        self.max_buffer_size = max_buffer_size
        self.max_incoming_transfer_size = max_incoming_transfer_size
        self.server_primary_trust_center = server_primary_trust_center
        self.server_backup_trust_center = server_backup_trust_center
        self.server_primary_binding_table_cache = server_primary_binding_table_cache
        self.server_backup_binding_table_cache = server_backup_binding_table_cache
        self.server_primary_discovery_cache = server_primary_discovery_cache
        self.server_backup_discovery_cache = server_backup_discovery_cache
        self.network_manager = network_manager
        self.stack_compliance_revision = stack_compliance_revision
        self.max_outgoing_transfer_size = max_outgoing_transfer_size
        self.extended_active_endpoint_list_available = extended_active_endpoint_list_available
        self.extended_simple_descriptors_list_available = extended_simple_descriptors_list_available

    def __repr__(self):
        return ("NodeDescriptor(" +
            "logical_type="+str(self.logical_type)+","+
            "complex_descriptor_available="+str(self.complex_descriptor_available)+","+
            "user_descriptor_available="+str(self.user_descriptor_available)+","+
            "aps_flags="+str(self.aps_flags)+","+
            "support_868_mhz="+str(self.support_868_mhz)+","+
            "support_902_mhz="+str(self.support_902_mhz)+","+
            "support_2400_mhz="+str(self.support_2400_mhz)+","+
            "alternate_pan_coordinator="+str(self.alternate_pan_coordinator)+","+
            "device_type="+str(self.device_type)+","+
            "power_source="+str(self.power_source)+","+
            "receiver_on_when_idle="+str(self.receiver_on_when_idle)+","+
            "security_capability="+str(self.security_capability)+","+
            "allocate_address="+str(self.allocate_address)+","+
            "manufacturer_code="+str(self.manufacturer_code)+","+
            "max_buffer_size="+str(self.max_buffer_size)+","+
            "max_incoming_transfer_size="+str(self.max_incoming_transfer_size)+","+
            "server_primary_trust_center="+str(self.server_primary_trust_center)+","+
            "server_backup_trust_center="+str(self.server_backup_trust_center)+","+
            "server_primary_binding_table_cache="+str(self.server_primary_binding_table_cache)+","+
            "server_backup_binding_table_cache="+str(self.server_backup_binding_table_cache)+","+
            "server_primary_discovery_cache="+str(self.server_primary_discovery_cache)+","+
            "server_backup_discovery_cache="+str(self.server_backup_discovery_cache)+","+
            "network_manager="+str(self.network_manager)+","+
            "stack_compliance_revision="+str(self.stack_compliance_revision)+","+
            "max_outgoing_transfer_size="+str(self.max_outgoing_transfer_size)+","+
            "extended_active_endpoint_list_available="+str(self.extended_active_endpoint_list_available)+","+
            "extended_simple_descriptors_list_available="+str(self.extended_simple_descriptors_list_available)+","+
        ")")

class SimpleDescriptor:
    """
    This class is a descriptor for an application.
    """
    def __init__(
                    self,
                    endpoint,
                    profile_identifier,
                    device_identifier,
                    device_version,
                    input_clusters,
                    output_clusters
    ):
        self.endpoint = endpoint
        self.profile_identifier = profile_identifier
        self.device_identifier = device_identifier
        self.device_version = device_version
        self.input_clusters = input_clusters
        self.output_clusters = output_clusters

    def __repr__(self):
        return (
            "SimpleDescriptor(" +
            "endpoint="+str(self.endpoint)+"," +
            "profile_identifier="+hex(self.profile_identifier)+"," +
            "device_identifier="+hex(self.device_identifier)+"," +
            "device_version="+hex(self.device_version)+"," +
            "input_clusters="+str(self.input_clusters)+"," +
            "output_clusters="+str(self.output_clusters) +
            ")"
        )



from whad.common.stack import LayerState, ContextualLayer, alias, source, state
from whad.scapy.layers.lorawan import MACPayloadUplink, MACPayloadDownlink

class LWMacLayerState(LayerState):

    def __init__(self):
        super().__init__()

        # Default state
        self.up_frame_counter = 0
        self.down_frame_counter = 0

@alias('mac')
@state(LWMacLayerState)
class LWMacLayer(ContextualLayer):
    
    def configure(self, options):
        # Initialize state
        pass

    def set_devaddr(self, dev_addr : int):
        """Save device address for debugging purpose.

        Set device network address (used for debugging).

        :param dev_addr: Device network address
        :type dev_addr: int
        """
        self.dev_addr = dev_addr

    def set_up_counter(self, value: int):
        """Set device uplink frame counter
        """
        self.state.up_frame_counter = value

    def set_down_counter(self, value : int):
        """Set device downlink frame counter

        :param value: Uplink frame counter value
        :type value: int
        """
        self.state.down_frame_counter = value

    def parse_mac_commands(self, commands : bytes):
        """Parse MAC commands, if any.

        Not supported yet, implementation to come.

        :param commands: MAC commands to process
        :type commands: bytes
        """
        pass

    @source('ll')
    def on_mac_frame(self, mac: MACPayloadUplink, confirmed : bool = False):
        """Handles uplink payload sent by link-layer.

        Payload is in clear as our link-layer has already decrypted and checked it.

        :param mac: MAC frame
        :type mac: MACPayloadUplink
        :param confirmed: uplink frame is from a confirmed uplink
        :type confirmed: bool
        """
        # Update our up frame counter
        self.state.up_frame_counter = mac.fcnt

        # Parse MAC commands from FOpts or payload depending
        # on FPort value.
        if mac.fport == 0:
            self.parse_mac_commands(bytes(mac.payload))
        else:
            if len(mac.fopts) > 0:
                self.parse_mac_commands(mac.fopts)
        
        # Notify data to gateway
        response = self.get_layer('ll').on_data_received(
            self.dev_addr,
            bytes(mac.payload),
            self.state.up_frame_counter
        )

        # Build response if required
        if response is not None:
            resp = MACPayloadDownlink(
                dev_addr = self.dev_addr,
                fcnt = self.state.down_frame_counter,
                fport=8 # data
            )/bytes(response)
            
            self.state.down_frame_counter += 1

            # Send back response to link-layer
            self.send('ll', resp)
        else:
            return None




"""Bluetooth Low Energy Advertising Data

This module provides classes corresponding to every supported Advertisement Record
(AD) that can be set in a device's advertising data or scan response data.

These classes are intended to be used when declaring a BLE peripheral device with
:class:`whad.ble.connector.peripheral.Peripheral`, but can also be helpful to
parse raw advertising data through class :class:`AdvDataFieldList`. 
"""
from struct import pack, unpack
from urllib.parse import urlparse
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.profile.attribute import UUID


class AdvDataError(Exception):
    """Advertisement Data error
    """

class AdvDataFieldListOverflow(Exception):
    """Advertisement data field list overflow
    """

class AdvDataField:
    """Advertisement basic data field.

    This class handles a basic advertisement data record (field) and its
    serialization.
    """

    def __init__(self, adv_type, value=b''):
        """Initialize an advertisement data record.

        :param int adv_type: Record type
        :param bytes value: Record value
        """
        self.__type = adv_type
        self.__value = value

    @property
    def type(self):
        """Return the record type
        """
        return self.__type
    
    def set_value(self, value):
        """Set advertising record value
        """
        self.__value = value

    def to_bytes(self):
        """Serialize record into a byte array

        :return: Serialized record
        :rtype: bytes
        """
        return pack('<BB', len(self.__value) + 1, self.__type) + self.__value


class AdvUuid16List(AdvDataField):
    """16-bit UUID list.

    This class provides a convenient way to handle AD records that store a list of
    16-bit UUIDs.
    """

    def __init__(self, eir_tag, *args):
        """Create a list of 16-bit UUIDs.

        :param int eir_tag: AD record type
        :param *args: Variable-length list of 16-bit UUIDs
        """
        # First, make sure all arguments are 16-bit UUID objects
        self.__uuids = []
        for arg in args:
            if not isinstance(arg, UUID):
                raise ValueError
            if arg.type != UUID.TYPE_16:
                raise ValueError
            # Append UUID
            self.__uuids.append(arg)

        # Pack data
        super().__init__(eir_tag, b''.join([uuid.packed for uuid in self.__uuids]))

    def __len__(self):
        return len(self.__uuids)

    def __getitem__(self, index):
        if 0 <= index < len(self.__uuids):
            return self.__uuids[index]

        # Not found
        raise IndexError

    @staticmethod
    def from_bytes(clazz, ad_record):
        """Deserialize a record containing a list of 16-bit UUIDs.

        :param class clazz: Class that will be instantiated (must inherit from AdvUuid16List)
        :param bytes ad_record: AD record to deserialize.
        """
        nb_uuids = int(len(ad_record)/2)
        uuids = []
        for i in range(nb_uuids):
            uuids.append(UUID(ad_record[i*2:(i+1)*2]))
        return clazz(*uuids)

class AdvUuid128List(AdvDataField):
    """128-bit UUID list.

    This class provides a convenient way to handle AD records that store a list of
    128-bit UUIDs.
    """

    def __init__(self, eir_tag, *args):
        """Create a list of 128-bit UUID.

        :param int eir_tag: AD record type
        :param *args: Variable-length list of 128-bit UUIDs
        """
        # First, make sure all arguments are 128-bit UUID objects
        self.__uuids = []
        for arg in args:
            if not isinstance(arg, UUID):
                raise ValueError
            if arg.type != UUID.TYPE_128:
                raise ValueError
            # Append UUID
            self.__uuids.append(arg)

        # Pack data
        super().__init__(eir_tag, b''.join([uuid.packed for uuid in self.__uuids]))

    def __len__(self):
        return len(self.__uuids)

    def __getitem__(self, index):
        if 0 <= index < len(self.__uuids):
            return self.__uuids[index]
        # Cannot find item.
        raise IndexError

    @staticmethod
    def from_bytes(clazz, ad_record):
        """Deserialize a record containing a list of 128-bit UUIDs.

        :param class clazz: Class that will be instantiated (must inherit from AdvUuid128List)
        :param bytes ad_record: AD record to deserialize.
        """
        nb_uuids = int(len(ad_record)/16)
        uuids = []
        for i in range(nb_uuids):
            uuids.append(UUID(ad_record[i*16:(i+1)*16]))
        return clazz(*uuids)


class AdvFlagsField(AdvDataField):
    """Advertisement Flags Data field

    This advertisement field specifies the device capabilities.
    """

    def __init__(self, limited_disc=False, general_disc=True, bredr_support=True,
                 le_bredr_support=False):
        """
        :param bool limited_disc: If set, enable the limited discoverable mode
        :param bool general_disc: If set, enable the generic discoverable mode
        :param bool bredr_support: If set, advertise the device does not support BR/EDR mode
        :param bool le_bredr_support: If set, advertise LE and BR/EDR support
        """
        flags = 0x00
        if limited_disc:
            flags |= 0x01
        if general_disc:
            flags |= 0x02
        if bredr_support:
            flags |= 0x04
        if le_bredr_support:
            flags |= 0x08
        super().__init__(0x01, bytes([flags]))

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvFlagsField AD record.

        :param bytes ad_record: AD record to deserialize
        :return: an AdvFlagsField object
        :rtype: AdvFlagsField 
        """

        if len(ad_record) != 1:
            raise AdvDataError

        limited_disc = (ad_record[0] & 0x01) != 0
        general_disc = (ad_record[0] & 0x02) != 0
        bredr_support = (ad_record[0] & 0x04) != 0
        lebredr_support = (ad_record[0] & 0x08) != 0
        return AdvFlagsField(
            limited_disc=limited_disc,
            general_disc=general_disc,
            bredr_support=bredr_support,
            le_bredr_support=lebredr_support
        )


class AdvShortenedLocalName(AdvDataField):
    """Device shortened local name data field

    This AD record (or field) contains the shortened name of the device that
    sends it.
    """

    def __init__(self, local_name):
        """Initialize this AdvShortenedLocalName object.

        :param str local_name: Shortened local name
        """
        self.__name = local_name
        super().__init__(0x08, local_name)

    @property
    def name(self):
        """Return this record shortened local name
        """
        return self.__name

    @name.setter
    def name(self, value: bytes):
        """Update shortened name
        """
        self.__name = value
        self.set_value(value)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvShortenedLocalName

        :param bytes ad_record: Serialized AdvShortenedLocalName AD record
        :return: An AdvShortenedLocalName object
        :rtype: AdvShortenedLocalName
        """
        return AdvShortenedLocalName(ad_record)


class AdvCompleteLocalName(AdvDataField):
    """Device complete local name
    
    This AD record (or field) contains the complete name of the device that
    sends it.
    """

    def __init__(self, local_name):
        self.__name = local_name
        super().__init__(0x09, local_name)

    @property
    def name(self):
        """Return the complete device name
        """
        return self.__name
    
    @name.setter
    def name(self, value: bytes):
        self.__name = value
        self.set_value(value)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvCompleteLocalName

        :param bytes ad_record: Serialized AdvCompleteLocalName AD record
        :return: An AdvCompleteLocalName object
        :rtype: AdvCompleteLocalName
        """
        return AdvCompleteLocalName(ad_record)


class AdvTxPowerLevel(AdvDataField):
    """Device Tx power level
    """

    def __init__(self, level):
        super().__init__(0x0A, bytes([level&0xff]))

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvTxPowerLevel

        :param bytes ad_record: Serialized AdvTxPowerLevel AD record
        :return: An AdvTxPowerLevel object
        :rtype: AdvTxPowerLevel
        """
        if len(ad_record) >= 1:
            return AdvTxPowerLevel(ad_record[0])
        # parsing error
        raise AdvDataError


class AdvManufacturerSpecificData(AdvDataField):
    """Device Manufacturer Specific Data
    """

    def __init__(self, company_id, data):
        super().__init__(0xFF, pack('<H', company_id&0xffff) + bytes(data))
        self.__company = company_id
        self.__data = data

    @property
    def company(self) -> int:
        """Company ID
        """
        return self.__company
    
    @company.setter
    def company(self, value):
        """Update company ID
        """
        self.__company = value
        self.set_value(pack('<H', value&0xffff) + bytes(self.__data))

    @property
    def data(self) -> bytes:
        """Manufacturer data
        """
        return self.__data
    
    @data.setter
    def data(self, value: bytes):
        """Update manufacturer data
        """
        self.__data = value
        self.set_value(pack('<H', self.__company&0xffff) + bytes(self.__data))

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvManufacturerSpecificData

        :param bytes ad_record: Serialized AdvManufacturerSpecificData AD record
        :return: An AdvManufacturerSpecificData object
        :rtype: AdvManufacturerSpecificData
        """
        if len(ad_record) >= 2:
            return AdvManufacturerSpecificData(
                unpack('<H', ad_record[:2])[0],
                ad_record[2:]
            )
        # Parsing error
        raise AdvDataError


class AdvIncServiceUuid16List(AdvUuid16List):
    """Incomplete list of 16-bit Service UUIDs
    """

    def __init__(self, *args):
        super().__init__(0x02, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvIncServiceUuid16List

        :param bytes ad_record: Serialized AdvIncServiceUuid16List AD record
        :return: An AdvIncServiceUuid16List object
        :rtype: AdvIncServiceUuid16List
        """
        return AdvUuid16List.from_bytes(AdvIncServiceUuid16List, ad_record)


class AdvCompServiceUuid16List(AdvUuid16List):
    """Complete list of 16-bit Service UUIDs
    """

    def __init__(self, *args):
        super().__init__(0x03, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvCompServiceUuid16List

        :param bytes ad_record: Serialized AdvCompServiceUuid16List AD record
        :return: An AdvCompServiceUuid16List object
        :rtype: AdvCompServiceUuid16List
        """
        return AdvUuid16List.from_bytes(AdvCompServiceUuid16List, ad_record)


class AdvIncServiceUuid128List(AdvUuid128List):
    """Incomplete list of 128-bit Service UUIDs
    """

    def __init__(self, *args):
        super().__init__(0x06, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvIncServiceUuid128List

        :param bytes ad_record: Serialized AdvIncServiceUuid128List AD record
        :return: An AdvIncServiceUuid128List object
        :rtype: AdvIncServiceUuid128List
        """
        return AdvUuid128List.from_bytes(AdvIncServiceUuid128List, ad_record)

class AdvCompServiceUuid128List(AdvUuid128List):
    """Complete list of 128-bit Service UUIDs
    """

    def __init__(self, *args):
        """Create a list of 128-bit UUID.
        """
        super().__init__(0x07, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvCompServiceUuid128List

        :param bytes ad_record: Serialized AdvCompServiceUuid128List AD record
        :return: An AdvCompServiceUuid128List object
        :rtype: AdvCompServiceUuid128List
        """
        return AdvUuid128List.from_bytes(AdvCompServiceUuid128List, ad_record)


class AdvSlaveConnIntervalRange(AdvDataField):
    """Advertising data Slave Connection Interval Range
    """

    def __init__(self, min_value=0xFFFF, max_value=0xFFFF):
        """Create a Slave Connection Interval Range advertising data record.

        :param int min_value: Min interval range (0x0006 - 0x0C80)
        :param int max_value: Max interval range (0x0006 - 0x0C80)
        """
        self.__range = [min_value, max_value]
        super().__init__(0x12, pack('<HH', min_value, max_value))

    @property
    def range(self):
        """Return the Slave connection interval range

        :rtype: list
        :return: Slave connection interval range
        """
        return self.__range

    @property
    def min(self):
        """Return the Slave connection minimal value.

        :rtype: int
        :return: Slave connection minimal value
        """
        return self.__range[0]

    @property
    def max(self):
        """Return the Slave connection maximal value.

        :rtype: int
        :return: Slave connection maximal value
        """
        return self.__range[1]

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvSlaveConnIntervalRange

        :param bytes ad_record: Serialized AdvSlaveConnIntervalRange AD record
        :return: An AdvSlaveConnIntervalRange object
        :rtype: AdvSlaveConnIntervalRange
        """
        if len(ad_record) == 4:
            min_value, max_value = unpack('<HH', ad_record)
            return AdvSlaveConnIntervalRange(min_value, max_value)
        # Parsing error
        raise AdvDataError


class AdvServiceSollicitationUuid16List(AdvUuid16List):
    """List of 16-bit Service sollicitation UUIDs.
    """

    def __init__(self, *args):
        """Create a list of 16-bit UUID.

        :param *args: Variable arguments list of 16-bit UUIDs
        """
        super().__init__(0x14, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvServiceSollicitationUuid16List

        :param bytes ad_record: Serialized AdvServiceSollicitationUuid16List AD record
        :return: An AdvServiceSollicitationUuid16List object
        :rtype: AdvServiceSollicitationUuid16List
        """
        return AdvUuid16List.from_bytes(AdvServiceSollicitationUuid16List, ad_record)


class AdvServiceSollicitationUuid128List(AdvUuid128List):
    """List of 128-bit Service sollicitation UUIDs.
    """

    def __init__(self, *args):
        """Create a list of 128-bit UUID.

        :param *args: Variable arguments list of 128-bit UUIDs
        """
        super().__init__(0x15, *args)

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvServiceSollicitationUuid128List

        :param bytes ad_record: Serialized AdvServiceSollicitationUuid128List AD record
        :return: An AdvServiceSollicitationUuid128List object
        :rtype: AdvServiceSollicitationUuid128List
        """
        return AdvUuid128List.from_bytes(AdvServiceSollicitationUuid128List, ad_record)


class AdvServiceData16(AdvDataField):
    """Service Data with 16-bit UUID.
    """

    def __init__(self, uuid, data):
        self.__uuid = uuid
        self.__data = data
        super().__init__(0x16, bytes(uuid.packed + data))

    @property
    def uuid(self):
        """Return Service 16-bit UUID

        :return: Service 16-bit UUID
        :rtype: UUID
        """
        return self.__uuid

    @property
    def data(self):
        """Returns Service data

        :return: Service data
        :rtype: bytes
        """
        return self.__data

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvServiceData16

        :param bytes ad_record: Serialized AdvServiceData16 AD record
        :return: An AdvServiceData16 object
        :rtype: AdvServiceData16
        """
        if len(ad_record) >= 2:
            uuid = UUID(unpack('<H', ad_record[:2])[0])
            data = ad_record[2:]
            return AdvServiceData16(uuid, data)
        # Parsing error
        raise AdvDataError


class AdvPublicTargetAddr(AdvDataField):
    """Public target address.
    """

    def __init__(self, *addresses):
        """Create a Public Target Address advertising data record.

        :param addresses: One or more :class:`BDAddress` objects.
        """
        # Parse given addresses
        self.__addresses = []
        for address in addresses:
            if isinstance(address, BDAddress):
                self.__addresses.append(address)
            elif isinstance(address, str):
                self.__addresses.append(BDAddress(address))

        super().__init__(0x17, b''.join([addr.value for addr in self.__addresses]))

    def __len__(self):
        return len(self.__addresses)

    def __getitem__(self, index):
        if 0 <= index < len(self.__addresses):
            return self.__addresses[index]
        # Parsing error.
        raise IndexError

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvPublicTargetAddr

        :param bytes ad_record: Serialized AdvPublicTargetAddr AD record
        :return: An AdvPublicTargetAddr object
        :rtype: AdvPublicTargetAddr
        """
        if len(ad_record) > 0 and ((len(ad_record) % 6) == 0):
            nb_addr = int(len(ad_record)/6)
            addresses = []
            for i in range(nb_addr):
                addresses.append(
                    BDAddress.from_bytes(ad_record[6*i:6*(i+1)])
                )
            return AdvPublicTargetAddr(*addresses)
        # Parsing error.
        raise AdvDataError


class AdvRandomTargetAddr(AdvDataField):
    """Random target address.
    """

    def __init__(self, *addresses):
        """Create a Public Target Address advertising data record.

        :param *addresses: One or more :class:`BDAddress` objects.
        """
        # Parse given addresses
        self.__addresses = []
        for address in addresses:
            if isinstance(address, BDAddress):
                self.__addresses.append(address)
            elif isinstance(address, str):
                self.__addresses.append(BDAddress(address))

        super().__init__(0x18, b''.join([addr.value for addr in self.__addresses]))

    def __len__(self):
        return len(self.__addresses)

    def __getitem__(self, index):
        if 0 <= index < len(self.__addresses):
            return self.__addresses[index]
        # Parsing error.
        raise IndexError

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvRandomTargetAddr

        :param bytes ad_record: Serialized AdvRandomTargetAddr AD record
        :return: An AdvRandomTargetAddr object
        :rtype: AdvRandomTargetAddr
        """
        if len(ad_record) > 0 and ((len(ad_record) % 6) == 0):
            nb_addr = int(len(ad_record)/6)
            addresses = []
            for i in range(nb_addr):
                addresses.append(
                    BDAddress.from_bytes(ad_record[6*i:6*(i+1)])
                )
            return AdvRandomTargetAddr(*addresses)
        # Parsing error.
        raise AdvDataError


class AdvAppearance(AdvDataField):
    """Device appearance advertising data record.
    """

    def __init__(self, appearance=0x0000):
        """Create an Device appearance advertisement record.

        :param int appearance: Device appearance (16-bit value)
        """
        if 0x0000 <= appearance <= 0xFFFF:
            self.__appearance = appearance
            super().__init__(0x19, pack('<H', appearance))
        else:
            raise AdvDataError

    @property
    def category(self):
        """Return device category

        :returns int: Device category
        """
        return self.__appearance >> 6

    @property
    def subcategory(self):
        """Return device sub-category

        :returns int: Device sub-category
        """
        return self.__appearance & 0x3f

    @staticmethod
    def from_bytes(ad_record):
        """Create an AdvAppearance object from serialized advertising data record.

        :param bytes ad_record: Serialized data record.
        :rtype: AdvAppearance
        :return: A new AdvAppearance object that represents the device appearance.
        """
        if len(ad_record) == 2:
            appearance = unpack('<H', ad_record)[0]
            return AdvAppearance(appearance)
        # Parsing error.
        raise AdvDataError


class AdvURI(AdvDataField):
    """Uniform Resource Identifier advertising data record.
    """

    SUPPORTED_SCHEMES = {
        'aaa'   : 0x0001,
        'aaas'  : 0x0002,
        'data'  : 0x000C,
        'ftp'   : 0x0011,
        'http'  : 0x0016,
        'https' : 0x0017,
        'mailto': 0x0026
    }

    def __init__(self, url):
        """Initialize an AdvURI object.

        :param str url: URL to include in the AdvURI AD record
        """
        url_info = urlparse(url)
        if url_info.scheme and url_info.scheme in AdvURI.SUPPORTED_SCHEMES:
            self.__scheme = url_info.scheme
            scheme = AdvURI.SUPPORTED_SCHEMES[url_info.scheme]
            self.__uri = url_info._replace(scheme='').geturl()
            encoded_uri = chr(scheme).encode("utf-8") + self.__uri.encode('utf-8')
            super().__init__(0x24, encoded_uri)
        else:
            raise AdvDataError

    @staticmethod
    def get_scheme(scheme_value):
        """Retrieve the associated scheme if known.
        """
        for s, value in AdvURI.SUPPORTED_SCHEMES.items():
            if value == scheme_value:
                return s
        return None

    @property
    def uri(self):
        """Return the URI

        :returns str: URI
        """
        return self.__uri

    @property
    def scheme(self):
        """Return the scheme

        :returns str: URL scheme
        """
        return self.__scheme

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvURI

        :param bytes ad_record: Serialized AdvURI AD record
        :return: An AdvURI object
        :rtype: AdvURI
        """
        if len(ad_record) >= 2:
            # Fetch the first UTF-8 character codepoint
            scheme = ord(ad_record.decode("utf-8")[0])
            scheme_size = len(ad_record.decode("utf-8")[0].encode("utf-8"))
            uri = ad_record[scheme_size:]
            scheme_alias = AdvURI.get_scheme(scheme)
            decoded_uri = uri.decode('utf-8')
            if scheme_alias is not None:
                scheme = AdvURI.get_scheme(scheme)
                return AdvURI(f"{scheme}:{decoded_uri}")

            return AdvURI(f"<0x{scheme:04x}>:{decoded_uri}")
        # Not enough data.
        return None


class AdvAdvertisingInterval(AdvDataField):
    """Advertising Interval record.
    """

    def __init__(self, interval):
        """Create an Advertising Interval record.

        :param int interval: Advertising interval record
        """
        if interval <= 0xFFFF:
            self.__interval = interval
            self.__interval_packed = pack('<H', interval)
        elif interval <= 0xFFFFFF:
            self.__interval = interval
            self.__interval_packed = bytes([
                interval & 0xFF,
                (interval & 0xFF00)>>8,
                (interval & 0xFF0000)>>16
            ])
        elif interval <= 0xFFFFFFFF:
            self.__interval = interval
            self.__interval_packed = pack('<I', interval)
        else:
            raise AdvDataError

        super().__init__(0x1A, self.__interval_packed)

    @property
    def interval(self):
        """Advertising interval
        """
        return self.__interval

    @staticmethod
    def from_bytes(ad_record):
        """Create an AdvAdvertisingInterval object from serialized record

        :param bytes ad_record: Serialized AdvAdvertisingInterval record
        :rtype: AdvAdvertisingInterval
        :return: An instance of AdvAdvertisingInterval
        """
        if len(ad_record) == 2:
            interval = unpack('<H', ad_record)[0]
            return AdvAdvertisingInterval(interval)
        if len(ad_record) == 3:
            interval = ad_record[0] | (ad_record[1]<<8) | (ad_record[2]<<16)
            return AdvAdvertisingInterval(interval)
        if len(ad_record) == 4:
            interval = unpack('<I', ad_record)[0]
            return AdvAdvertisingInterval(interval)
        # Parsing error
        raise AdvDataError

class AdvBluetoothDeviceAddr(AdvDataField):
    """Bluetooth Device Address information record.
    """

    def __init__(self, bd_address, public=False):
        """Create a Bluetooth Device Address advertising data record

        If both `random` and  `public` are set to True, BD address
        will be considered as `public`.

        :param bd_address: Bluetooth Device Address
        :type bd_address: str, BDAddress
        :param bool public: Set to True to specify a public BD address, False to
                            specify a random BD address
        """
        self.__public = public

        if isinstance(bd_address, BDAddress):
            address = bd_address
        elif isinstance(bd_address, str):
            address = BDAddress(bd_address)
        else:
            raise AdvDataError

        if public:
            suffix = bytes([0x00])
        else:
            suffix = bytes([0x01])

        super().__init__(0x1B,address.value + suffix)

    @property
    def is_public(self):
        """Determine if the bluetooth device address is public

        :return bool: True if address is public, False otherwise
        """
        return self.__public

    @property
    def is_random(self):
        """Determine if the bluetooth device address is random

        :return bool: True if address is random, False otherwise
        """
        return not self.__public


    @staticmethod
    def from_bytes(ad_record):
        """Create an AdvBluetoothDeviceAddr object from serialized record

        :param bytes ad_record: Serialized AdvBluetoothDeviceAddr record
        :rtype: AdvBluetoothDeviceAddr
        :return: An instance of AdvBluetoothDeviceAddr
        """
        if len(ad_record) == 7:
            public = ad_record[6] == 0x00
            address = BDAddress.from_bytes(ad_record[:6])
            return AdvBluetoothDeviceAddr(address, public=public)
        # Parsing error
        raise AdvDataError


class AdvLeRole(AdvDataField):
    """LE Role advertising data record.
    """

    ONLY_PERIPHERAL_ROLE = 0x00
    ONLY_CENTRAL_ROLE = 0x01
    PREFERRED_PERIPHERAL_ROLE = 0x02
    PREFERRED_CENTRAL_ROLE = 0x03


    def __init__(self, role):
        """Create an AdvLeRole advertising data record.

        :param int role: LE role
        """
        self.__role = role
        if self.__role >= 0 and self.__role < 4:
            super().__init__(0x1C, pack('<B', role))
        else:
            raise AdvDataError

    @property
    def role(self):
        """Return LE role value
        """
        return self.__role

    @staticmethod
    def from_bytes(ad_record):
        """Create an AdvLeRole object from serialized record

        :param bytes ad_record: Serialized AdvLeRole record
        :rtype: AdvLeRole
        :return: An instance of AdvLeRole
        """
        if len(ad_record) == 1:
            role = ad_record[0]
            return AdvLeRole(role)
        # Parsing error
        raise AdvDataError


class AdvServiceDataUuid128(AdvDataField):
    """Service Data with 128-bit UUID.
    """

    def __init__(self, uuid, data):
        """Initialize a 128-bit UUID Service data

        :param UUID uuid: Service 128-bit UUID
        :param bytes data: Service associated data
        """
        self.__uuid = uuid
        if self.__uuid.type == UUID.TYPE_128:
            self.__data = data
            super().__init__(0x21, bytes(uuid.packed + data))
        else:
            raise AdvDataError

    @property
    def uuid(self):
        """Returns service's 128-bit UUID
        """
        return self.__uuid

    @property
    def data(self):
        """Returns service's data
        """
        return self.__data

    @staticmethod
    def from_bytes(ad_record):
        """Create an AdvServiceDataUuid128 object from serialized record

        :param bytes ad_record: Serialized AdvServiceDataUuid128 record
        :rtype: AdvServiceDataUuid128
        :return: An instance of AdvServiceDataUuid128
        """
        if len(ad_record) >= 16:
            uuid = UUID(ad_record[:16])
            data = ad_record[16:]
            return AdvServiceDataUuid128(uuid, data)
        # Parsing error
        raise AdvDataError


class AdvLeSupportedFeatures(AdvDataField):
    """LE Supported Features
    """

    def __init__(self, encryption=False, conn_param_update=False, ext_reject_ind=False,
                 slave_features_exchange=False, ping=False, data_packet_length=False,
                 privacy=False, ext_scanner_filter_policies=False):
        """Initialize an AdvLeSupportedFeatures AD record

        :param bool encryption: True if LE encryption is supported, False otherwise
        :param bool conn_param_update: True if connection parameter update request
                                       procedure is supported, False otherwise
        :param bool ext_reject_ind: True if extended rejection is supported, False otherwise
        :param bool slave_features_exchange: True if slave-initiated features exchange is
                                             supported, False otherwise
        :param bool slave_features_exchange: True if LE ping procedure is supported, False otherwise
        :param bool data_packet_length: True if LE data packet length procedure is supported,
                                        False otherwise
        :param bool privacy: True if privacy feature is supported, False otherwise
        :param bool ext_scanner_filter_policies: True if extended scanner filtering
                                                 policies are supported, False otherwise
        """
        # Save parameters
        self.__encryption = encryption
        self.__conn_param_update = conn_param_update
        self.__ext_reject_ind = ext_reject_ind
        self.__slave_features_exchange = slave_features_exchange
        self.__ping = ping
        self.__data_packet_length = data_packet_length
        self.__privacy = privacy
        self.__ext_scanner_filter_policies = ext_scanner_filter_policies

        # Generate bitmap
        features = 0x00
        if self.__encryption:
            features |= 1
        if self.__conn_param_update:
            features |= (1 << 1)
        if self.__ext_reject_ind:
            features |= (1 << 2)
        if self.__slave_features_exchange:
            features |= (1 << 3)
        if self.__ping:
            features |= (1 << 4)
        if self.__data_packet_length:
            features |= (1 << 5)
        if self.__privacy:
            features |= (1 << 6)
        if self.__ext_scanner_filter_policies:
            features |= (1 << 7)

        # Since only 8 bits are used in version 5.3, we code this on a single byte
        super().__init__(0x27, pack('<B', features))

    @property
    def has_encryption(self):
        """Determine if encryption is supported.
        """
        return self.__encryption

    @property
    def has_conn_param_update(self):
        """Determine if connection parameters update procedure is supported.
        """
        return self.__conn_param_update

    @property
    def has_ext_reject_ind(self):
        """Determine if extended reject_ind is supported.
        """
        return self.__ext_reject_ind

    @property
    def has_slave_features_exchange(self):
        """Determine if the slave features exchange procedure is supported.
        """
        return self.__slave_features_exchange

    @property
    def has_ping(self):
        """Determine if GATT ping procedure is supported.
        """
        return self.__ping

    @property
    def has_data_packet_length(self):
        """Determine if packet length is supported.
        """
        return self.__data_packet_length

    @property
    def has_privacy(self):
        """Determine if device supports privacy features.
        """
        return self.__privacy

    @property
    def has_ext_scanner_filter_policies(self):
        """Determine if device supports extended scanning
        """
        return self.__ext_scanner_filter_policies

    @staticmethod
    def from_bytes(ad_record):
        """Deserialize an AdvDataError

        :param bytes ad_record: Serialized AdvDataError AD record
        :return: An AdvDataError object
        :rtype: AdvDataError
        """
        if len(ad_record) >= 1:
            # Parse features set
            features = 0x00
            for i, record in enumerate(ad_record):
                features |= (record << (8*i))

            # Deduce our flags
            encryption = (features & 1) != 0
            conn_param_update = (features & (1<<1)) != 0
            ext_reject_ind = (features & (1<<2)) != 0
            slave_features_exchange = (features & (1<<3)) != 0
            ping = (features & (1<<4)) != 0
            data_packet_length = (features & (1<<5)) != 0
            privacy = (features & (1<<6)) != 0
            ext_scanner_filter_policies = (features & (1<<7)) != 0
            return AdvLeSupportedFeatures(
                encryption=encryption,
                conn_param_update=conn_param_update,
                ext_reject_ind=ext_reject_ind,
                slave_features_exchange=slave_features_exchange,
                ping=ping,
                data_packet_length=data_packet_length,
                privacy=privacy,
                ext_scanner_filter_policies=ext_scanner_filter_policies
            )

        # Parsing error
        raise AdvDataError


class EddystoneUrl(AdvServiceData16):
    """Eddystone-URL advertising data record.

    No deserialization routine is provided for this record (yet).
    """

    SCHEMES = [
        "http://www.",
        "https://www.",
        "http://",
        "https://",
    ]

    EXTENSIONS = [
        ".com/", ".org/", ".edu/", ".net/", ".info/", ".biz/", ".gov/",
        ".com", ".org", ".edu", ".net", ".info", ".biz", ".gov",
    ]

    def encode_url(self, url):
        """Encode an URL
        """
        i = 0
        data = []

        for s, scheme in enumerate(self.SCHEMES):
            if url.startswith(scheme):
                data.append(s)
                i += len(scheme)
                break
        else:
            raise AdvDataError

        while i < len(url):
            if url[i] == '.':
                for e, expansion in enumerate(self.EXTENSIONS):
                    if url.startswith(expansion, i):
                        data.append(e)
                        i += len(expansion)
                        break
                else:
                    data.append(0x2E)
                    i += 1
            else:
                data.append(ord(url[i]))
                i += 1
        return data

    def __init__(self, url):
        """Create an EddystoneUrl advertising data record (Custom service data record).

        :param str url: Url to embed into this Eddystone-URL record.
        """
        eddystone_url = self.encode_url(url)
        super().__init__(UUID(0xFEAA), bytes([0x10, 0xF8] + eddystone_url))


class AdvDataFieldList:
    """Advertisement field list

    This class provides a convenient way to manage BLE advertisement records
    as well as the serialization/deserialization of such data.
    """

    EIR_HANDLERS = {
        0x01: AdvFlagsField,
        0x02: AdvIncServiceUuid16List,
        0x03: AdvCompServiceUuid16List,
        #0x04: AdvIncServiceUUID32List,
        #0x05: AdvCompServiceUUID32List,
        0x06: AdvIncServiceUuid128List,
        0x07: AdvCompServiceUuid128List,
        0x08: AdvShortenedLocalName,
        0x09: AdvCompleteLocalName,
        0x0A: AdvTxPowerLevel,
        #0x0D: AdvDeviceClass,
        #0x10: AdvDeviceId,
        0x12: AdvSlaveConnIntervalRange,
        0x14: AdvServiceSollicitationUuid16List,
        0x15: AdvServiceSollicitationUuid128List,
        0x16: AdvServiceData16,
        0x17: AdvPublicTargetAddr,
        0x18: AdvRandomTargetAddr,
        0x19: AdvAppearance,
        0x1A: AdvAdvertisingInterval,
        0x1B: AdvBluetoothDeviceAddr,
        0x1C: AdvLeRole,
        #0x1F: AdvServiceSollicitationUuid32List,
        #0x20: AdvServiceDataUuid32,
        0x21: AdvServiceDataUuid128,
        0x24: AdvURI,
        #0x25: AdvIndoorPositioning,
        #0x26: AdvTransportDiscData,
        0x27: AdvLeSupportedFeatures,
        #0x28: AdvChannelMapUpdateIndication,
        #0x29: AdvPbAdv,
        #0x2A: AdvMeshMessage,
        #0x2B: AdvMeshBeacon,
        #0x2C: AdvBigInfo,
        #0x2D: AdvBroadcastCode,
        #0x2F: AdvAdvertisingIntervalLong,
        #0x30: AdvBroadcastName,
        #0x3D: Adv3dInfoData,
        0xFF: AdvManufacturerSpecificData
    }


    def __init__(self, *args):
        self.__fields = []
        for field in args:
            self.add(field)

    def __len__(self):
        return len(self.__fields)

    def __getitem__(self, index):
        if 0 <= index < len(self.__fields):
            return self.__fields[index]
        # Error.
        raise IndexError

    def add(self, item):
        """Add item to advertising data
        """
        if isinstance(item, AdvDataField):
            self.__fields.append(item)
        else:
            raise AttributeError

    def get(self, adv_type) -> AdvDataField:
        """Find the first advertising record of the specified type
        """
        for field in self.__fields:
            if isinstance(field, adv_type):
                return field
        return None
    
    def remove(self, field: AdvDataField) -> bool:
        """Remove a record from this list
        """
        for f in self.__fields:
            if f == field:
                self.__fields.remove(f)
                return True
        return False

    def to_bytes(self):
        """Convert field list to bytes

        :return bytes: Serialized AD records list
        """
        output = b''
        for field in self.__fields:
            field_record = field.to_bytes()
            if len(output) + len(field_record) <= 31:
                output += field_record
            else:
                raise AdvDataFieldListOverflow
        return output

    @staticmethod
    def from_bytes(adv_data):
        """Convert raw advertising data into an AdvDataFieldList object.

        :param bytes adv_data: Raw advertising data
        :rtype: AdvDataFieldList
        :return: Instance of AdvDataFieldList populated with records
        """
        if len(adv_data) > 31:
            raise AdvDataFieldListOverflow

        adv_list = AdvDataFieldList()
        while len(adv_data) >= 2:
            # Unpack length and eir_tag
            length, eir_tag = unpack('<BB', adv_data[:2])

            # Check length
            if len(adv_data[2:]) >= (length - 1):
                # Extract EIR data
                eir_payload = adv_data[2:2 + length - 1]

                # Add record based on EIR tag
                if eir_tag in AdvDataFieldList.EIR_HANDLERS:
                    adv_list.add(
                        AdvDataFieldList.EIR_HANDLERS[eir_tag].from_bytes(eir_payload)
                    )
                adv_data = adv_data[length+1:]
            else:
                raise AdvDataError
        return adv_list
